"""
Support for hash calculations
"""

from __future__ import annotations

import hashlib
from pathlib import Path


def get_file_hash(
    fp: Path, algorithm: str = "sha256", buffer_size: int = 64 * 1024**2
) -> str:
    """
    Get the hash of a file

    Parameters
    ----------
    fp
        Path to the file

    algorithm
        Algorithm to use during hashing

    buffer_size
        The size of the chunks to use when hashing the file

    Returns
    -------
    :
        Hash of the file
    """
    hasher = hashlib.new(algorithm)
    with open(fp, "rb") as fh:
        data = fh.read(buffer_size)
        while data:
            hasher.update(data)
            data = fh.read(buffer_size)

    return hasher.hexdigest()



"""
AR6 components
"""

from __future__ import annotations

from .harmonisation import AR6Harmoniser
from .infilling import AR6Infiller, get_ar6_full_historical_emissions
from .post_processing import AR6PostProcessor
from .pre_processing import AR6PreProcessor
from .scm_running import AR6SCMRunner

# from .workflow import run_ar6_workflow

__all__ = [
    "AR6Harmoniser",
    "AR6Infiller",
    "AR6PostProcessor",
    "AR6PreProcessor",
    "AR6SCMRunner",
    # "run_ar6_workflow",
    "get_ar6_full_historical_emissions",
]



"""
CMIP7 ScenarioMIP components
"""

from __future__ import annotations

from gcages.cmip7_scenariomip.pre_processing import (
    CMIP7ScenarioMIPPreProcessingResult,
    CMIP7ScenarioMIPPreProcessor,
    ReaggregatorBasic,
    ReaggregatorLike,
)

__all__ = [
    "CMIP7ScenarioMIPPreProcessingResult",
    "CMIP7ScenarioMIPPreProcessor",
    "ReaggregatorBasic",
    "ReaggregatorLike",
]



"""
Emissions variables database
"""

from __future__ import annotations

import pandas as pd

EMISSIONS_VARIABLES = pd.DataFrame(
    [
        (
            "Emissions|BC",
            "Emissions|BC",
            "Emissions|BC",
            "Emissions|BC",
            "Emissions|BC",
            "Emissions|BC",
        ),
        (
            "Emissions|C2F6",
            "Emissions|C2F6",
            "Emissions|PFC|C2F6",
            "Emissions|C2F6",
            "Emissions|F-Gases|PFC|C2F6",
            "Emissions|PFC|C2F6",
        ),
        (
            "Emissions|C3F8",
            "Emissions|C3F8",
            "Emissions|PFC|C3F8",
            "Emissions|C3F8",
            "Emissions|F-Gases|PFC|C3F8",
            "Emissions|PFC|C3F8",
        ),
        (
            "Emissions|C4F10",
            "Emissions|C4F10",
            "Emissions|PFC|C4F10",
            "Emissions|C4F10",
            "Emissions|F-Gases|PFC|C4F10",
            "Emissions|PFC|C4F10",
        ),
        (
            "Emissions|C5F12",
            "Emissions|C5F12",
            "Emissions|PFC|C5F12",
            "Emissions|C5F12",
            "Emissions|F-Gases|PFC|C5F12",
            "Emissions|PFC|C5F12",
        ),
        (
            "Emissions|C6F14",
            "Emissions|C6F14",
            "Emissions|PFC|C6F14",
            "Emissions|C6F14",
            "Emissions|F-Gases|PFC|C6F14",
            "Emissions|PFC|C6F14",
        ),
        (
            "Emissions|C7F16",
            "Emissions|C7F16",
            "Emissions|PFC|C7F16",
            "Emissions|C7F16",
            "Emissions|F-Gases|PFC|C7F16",
            "Emissions|PFC|C7F16",
        ),
        (
            "Emissions|C8F18",
            "Emissions|C8F18",
            "Emissions|PFC|C8F18",
            "Emissions|C8F18",
            "Emissions|F-Gases|PFC|C8F18",
            "Emissions|PFC|C8F18",
        ),
        (
            "Emissions|CF4",
            "Emissions|CF4",
            "Emissions|PFC|CF4",
            "Emissions|CF4",
            "Emissions|F-Gases|PFC|CF4",
            "Emissions|PFC|CF4",
        ),
        (
            "Emissions|CH4",
            "Emissions|CH4",
            "Emissions|CH4",
            "Emissions|CH4",
            "Emissions|CH4",
            "Emissions|CH4",
        ),
        (
            "Emissions|CO",
            "Emissions|CO",
            "Emissions|CO",
            "Emissions|CO",
            "Emissions|CO",
            "Emissions|CO",
        ),
        (
            "Emissions|CO2",
            "Emissions|CO2",
            "Emissions|CO2",
            "Emissions|CO2",
            "Emissions|CO2",
            "Emissions|CO2",
        ),
        (
            "Emissions|CO2|Biosphere",
            "Emissions|CO2|MAGICC AFOLU",
            "Emissions|CO2|AFOLU",
            "Emissions|CO2|AFOLU",
            "Emissions|CO2|MAGICC AFOLU",
            "Emissions|CO2|AFOLU",
        ),
        (
            "Emissions|CO2|Fossil",
            "Emissions|CO2|MAGICC Fossil and Industrial",
            "Emissions|CO2|Energy and Industrial Processes",
            "Emissions|CO2|Energy and Industrial Processes",
            "Emissions|CO2|MAGICC Fossil and Industrial",
            "Emissions|CO2|Energy and Industrial Processes",
        ),
        (
            "Emissions|HFC125",
            "Emissions|HFC125",
            "Emissions|HFC|HFC125",
            "Emissions|HFC|HFC125",
            "Emissions|F-Gases|HFC|HFC125",
            "Emissions|HFC|HFC125",
        ),
        (
            "Emissions|HFC134a",
            "Emissions|HFC134a",
            "Emissions|HFC|HFC134a",
            "Emissions|HFC|HFC134a",
            "Emissions|F-Gases|HFC|HFC134a",
            "Emissions|HFC|HFC134a",
        ),
        (
            "Emissions|HFC143a",
            "Emissions|HFC143a",
            "Emissions|HFC|HFC143a",
            "Emissions|HFC|HFC143a",
            "Emissions|F-Gases|HFC|HFC143a",
            "Emissions|HFC|HFC143a",
        ),
        (
            "Emissions|HFC152a",
            "Emissions|HFC152a",
            "Emissions|HFC|HFC152a",
            "Emissions|HFC|HFC152a",
            "Emissions|F-Gases|HFC|HFC152a",
            "Emissions|HFC|HFC152a",
        ),
        (
            "Emissions|HFC227ea",
            "Emissions|HFC227ea",
            "Emissions|HFC|HFC227ea",
            "Emissions|HFC|HFC227ea",
            "Emissions|F-Gases|HFC|HFC227ea",
            "Emissions|HFC|HFC227ea",
        ),
        (
            "Emissions|HFC23",
            "Emissions|HFC23",
            "Emissions|HFC|HFC23",
            "Emissions|HFC|HFC23",
            "Emissions|F-Gases|HFC|HFC23",
            "Emissions|HFC|HFC23",
        ),
        (
            "Emissions|HFC236fa",
            "Emissions|HFC236fa",
            "Emissions|HFC|HFC236fa",
            "Emissions|HFC|HFC236fa",
            "Emissions|F-Gases|HFC|HFC236fa",
            "Emissions|HFC|HFC236fa",
        ),
        # No idea what is going on here, anyway
        (
            "Emissions|HFC245fa",
            "Emissions|HFC245fa",
            "Emissions|HFC|HFC245ca",
            "Emissions|HFC|HFC245fa",
            "Emissions|F-Gases|HFC|HFC245fa",
            "Emissions|HFC|HFC245fa",
        ),
        (
            "Emissions|HFC32",
            "Emissions|HFC32",
            "Emissions|HFC|HFC32",
            "Emissions|HFC|HFC32",
            "Emissions|F-Gases|HFC|HFC32",
            "Emissions|HFC|HFC32",
        ),
        (
            "Emissions|HFC365mfc",
            "Emissions|HFC365mfc",
            "Emissions|HFC|HFC365mfc",
            "Emissions|HFC|HFC365mfc",
            "Emissions|F-Gases|HFC|HFC365mfc",
            "Emissions|HFC|HFC365mfc",
        ),
        (
            "Emissions|HFC4310mee",
            "Emissions|HFC4310mee",
            "Emissions|HFC|HFC43-10",
            "Emissions|HFC|HFC43-10",
            "Emissions|F-Gases|HFC|HFC4310mee",
            "Emissions|HFC|HFC43-10",
        ),
        (
            "Emissions|CCl4",
            "Emissions|CCl4",
            "Emissions|CCl4",
            "Emissions|CCl4",
            "Emissions|Montreal Gases|CCl4",
            "Emissions|CCl4",
        ),
        (
            "Emissions|CFC11",
            "Emissions|CFC11",
            "Emissions|CFC11",
            "Emissions|CFC11",
            "Emissions|Montreal Gases|CFC|CFC11",
            "Emissions|CFC11",
        ),
        (
            "Emissions|CFC113",
            "Emissions|CFC113",
            "Emissions|CFC113",
            "Emissions|CFC113",
            "Emissions|Montreal Gases|CFC|CFC113",
            "Emissions|CFC113",
        ),
        (
            "Emissions|CFC114",
            "Emissions|CFC114",
            "Emissions|CFC114",
            "Emissions|CFC114",
            "Emissions|Montreal Gases|CFC|CFC114",
            "Emissions|CFC114",
        ),
        (
            "Emissions|CFC115",
            "Emissions|CFC115",
            "Emissions|CFC115",
            "Emissions|CFC115",
            "Emissions|Montreal Gases|CFC|CFC115",
            "Emissions|CFC115",
        ),
        (
            "Emissions|CFC12",
            "Emissions|CFC12",
            "Emissions|CFC12",
            "Emissions|CFC12",
            "Emissions|Montreal Gases|CFC|CFC12",
            "Emissions|CFC12",
        ),
        (
            "Emissions|CH2Cl2",
            "Emissions|CH2Cl2",
            "Emissions|CH2Cl2",
            "Emissions|CH2Cl2",
            "Emissions|Montreal Gases|CH2Cl2",
            "Emissions|CH2Cl2",
        ),
        (
            "Emissions|CH3Br",
            "Emissions|CH3Br",
            "Emissions|CH3Br",
            "Emissions|CH3Br",
            "Emissions|Montreal Gases|CH3Br",
            "Emissions|CH3Br",
        ),
        (
            "Emissions|CH3CCl3",
            "Emissions|CH3CCl3",
            "Emissions|CH3CCl3",
            "Emissions|CH3CCl3",
            "Emissions|Montreal Gases|CH3CCl3",
            "Emissions|CH3CCl3",
        ),
        (
            "Emissions|CH3Cl",
            "Emissions|CH3Cl",
            "Emissions|CH3Cl",
            "Emissions|CH3Cl",
            "Emissions|Montreal Gases|CH3Cl",
            "Emissions|CH3Cl",
        ),
        (
            "Emissions|CHCl3",
            "Emissions|CHCl3",
            "Emissions|CHCl3",
            "Emissions|CHCl3",
            "Emissions|Montreal Gases|CHCl3",
            "Emissions|CHCl3",
        ),
        (
            "Emissions|HCFC141b",
            "Emissions|HCFC141b",
            "Emissions|HCFC141b",
            "Emissions|HCFC141b",
            "Emissions|Montreal Gases|HCFC141b",
            "Emissions|HCFC141b",
        ),
        (
            "Emissions|HCFC142b",
            "Emissions|HCFC142b",
            "Emissions|HCFC142b",
            "Emissions|HCFC142b",
            "Emissions|Montreal Gases|HCFC142b",
            "Emissions|HCFC142b",
        ),
        (
            "Emissions|HCFC22",
            "Emissions|HCFC22",
            "Emissions|HCFC22",
            "Emissions|HCFC22",
            "Emissions|Montreal Gases|HCFC22",
            "Emissions|HCFC22",
        ),
        (
            "Emissions|Halon1202",
            "Emissions|Halon1202",
            "Emissions|Halon1202",
            "Emissions|Halon1202",
            "Emissions|Montreal Gases|Halon1202",
            "Emissions|Halon1202",
        ),
        (
            "Emissions|Halon1211",
            "Emissions|Halon1211",
            "Emissions|Halon1211",
            "Emissions|Halon1211",
            "Emissions|Montreal Gases|Halon1211",
            "Emissions|Halon1211",
        ),
        (
            "Emissions|Halon1301",
            "Emissions|Halon1301",
            "Emissions|Halon1301",
            "Emissions|Halon1301",
            "Emissions|Montreal Gases|Halon1301",
            "Emissions|Halon1301",
        ),
        (
            "Emissions|Halon2402",
            "Emissions|Halon2402",
            "Emissions|Halon2402",
            "Emissions|Halon2402",
            "Emissions|Montreal Gases|Halon2402",
            "Emissions|Halon2402",
        ),
        (
            "Emissions|N2O",
            "Emissions|N2O",
            "Emissions|N2O",
            "Emissions|N2O",
            "Emissions|N2O",
            "Emissions|N2O",
        ),
        (
            "Emissions|NF3",
            "Emissions|NF3",
            "Emissions|NF3",
            "Emissions|NF3",
            "Emissions|F-Gases|NF3",
            "Emissions|NF3",
        ),
        (
            "Emissions|NH3",
            "Emissions|NH3",
            "Emissions|NH3",
            "Emissions|NH3",
            "Emissions|NH3",
            "Emissions|NH3",
        ),
        (
            "Emissions|NOx",
            "Emissions|NOx",
            "Emissions|NOx",
            "Emissions|NOx",
            "Emissions|NOx",
            "Emissions|NOx",
        ),
        (
            "Emissions|OC",
            "Emissions|OC",
            "Emissions|OC",
            "Emissions|OC",
            "Emissions|OC",
            "Emissions|OC",
        ),
        (
            "Emissions|SF6",
            "Emissions|SF6",
            "Emissions|SF6",
            "Emissions|SF6",
            "Emissions|F-Gases|SF6",
            "Emissions|SF6",
        ),
        (
            "Emissions|SO2F2",
            "Emissions|SO2F2",
            "Emissions|SO2F2",
            "Emissions|SO2F2",
            "Emissions|F-Gases|SO2F2",
            "Emissions|SO2F2",
        ),
        (
            "Emissions|SOx",
            "Emissions|Sulfur",
            "Emissions|Sulfur",
            "Emissions|Sulfur",
            "Emissions|Sulfur",
            "Emissions|Sulfur",
        ),
        (
            "Emissions|NMVOC",
            "Emissions|VOC",
            "Emissions|VOC",
            "Emissions|VOC",
            "Emissions|VOC",
            "Emissions|VOC",
        ),
        (
            "Emissions|cC4F8",
            "Emissions|cC4F8",
            "Emissions|PFC|cC4F8",
            "Emissions|cC4F8",
            "Emissions|F-Gases|PFC|cC4F8",
            "Emissions|PFC|cC4F8",
        ),
    ],
    columns=[
        "gcages",
        "openscm_runner",
        "iamc",
        "cmip7_scenariomip",
        "rcmip",
        "ar6_cfc_infilling_db",
    ],
)
"""
Database of emissions variables names according to different naming schemes

You will likely not need to access this variable directly,
and instead will use [convert_variable_name][gcages.renaming.].
"""



"""
Simple climate model (SCM) running part of the AR6 workflow
"""

from __future__ import annotations

import json
import multiprocessing
import os
from functools import partial
from pathlib import Path
from typing import Any

import numpy as np
import pandas as pd
from attrs import define, field
from pandas_openscm.db import OpenSCMDB
from pandas_openscm.index_manipulation import update_index_levels_func

from gcages.assertions import (
    assert_data_is_all_numeric,
    assert_has_data_for_times,
    assert_has_index_levels,
    assert_index_is_multiindex,
)
from gcages.completeness import assert_all_groups_are_complete
from gcages.exceptions import MissingOptionalDependencyError
from gcages.harmonisation import assert_harmonised
from gcages.hashing import get_file_hash
from gcages.renaming import SupportedNamingConventions, convert_variable_name
from gcages.scm_running import (
    convert_openscm_runner_output_names_to_magicc_output_names,
    run_scms,
)
from gcages.units_helpers import assert_has_no_pint_incompatible_characters

DEFAULT_OUTPUT_VARIABLES: tuple[str, ...] = (
    # GSAT
    "Surface Air Temperature Change",
    # GMST
    "Surface Air Ocean Blended Temperature Change",
    # ERFs
    "Effective Radiative Forcing",
    "Effective Radiative Forcing|Anthropogenic",
    "Effective Radiative Forcing|Aerosols",
    "Effective Radiative Forcing|Aerosols|Direct Effect",
    "Effective Radiative Forcing|Aerosols|Direct Effect|BC",
    "Effective Radiative Forcing|Aerosols|Direct Effect|OC",
    "Effective Radiative Forcing|Aerosols|Direct Effect|SOx",
    "Effective Radiative Forcing|Aerosols|Indirect Effect",
    "Effective Radiative Forcing|Greenhouse Gases",
    "Effective Radiative Forcing|CO2",
    "Effective Radiative Forcing|CH4",
    "Effective Radiative Forcing|N2O",
    "Effective Radiative Forcing|F-Gases",
    "Effective Radiative Forcing|Montreal Protocol Halogen Gases",
    "Effective Radiative Forcing|Ozone",
    # Heat uptake
    "Heat Uptake",
    # "Heat Uptake|Ocean",
    # Atmospheric concentrations
    "Atmospheric Concentrations|CO2",
    "Atmospheric Concentrations|CH4",
    "Atmospheric Concentrations|N2O",
    # Carbon cycle
    "Net Atmosphere to Land Flux|CO2",
    "Net Atmosphere to Ocean Flux|CO2",
    # Permafrost
    "Net Land to Atmosphere Flux|CO2|Earth System Feedbacks|Permafrost",
    "Net Land to Atmosphere Flux|CH4|Earth System Feedbacks|Permafrost",
)
"""
Default output variables

Note that it can be a bit of work
to get these variables to actually appear in the output,
depending on which simple climate model you're using.
"""


def check_ar6_magicc7_version() -> None:
    """
    Check that the MAGICC7 version is what was used in AR6
    """
    try:
        import openscm_runner.adapters
    except ImportError as exc:
        raise MissingOptionalDependencyError(
            "check_ar6_magicc7_version", requirement="openscm_runner"
        ) from exc

    if openscm_runner.adapters.MAGICC7.get_version() != "v7.5.3":  # type: ignore
        raise AssertionError(openscm_runner.adapters.MAGICC7.get_version())  # type: ignore


def load_ar6_magicc_probabilistic_config(filepath: Path) -> list[dict[str, Any]]:
    """
    Load the probabilistic config used with MAGICC in AR6

    Parameters
    ----------
    filepath
        Filepath from which to load the probabilistic configuration

    Returns
    -------
    :
        Probabilistic configuration used with MAGICC in AR6

    Raises
    ------
    AssertionError
        `filepath` points to a file that does not have the expected hash
    """
    fp_hash = get_file_hash(filepath, algorithm="sha256")
    fp_hash_exp = "f4481549e2309e3f32de9095bcfc1adc531b8ce985201690fd889d49def8a02f"
    if fp_hash != fp_hash_exp:
        msg = (
            f"The sha256 hash of {filepath} is {fp_hash}. "
            f"This does not match what we expect ({fp_hash_exp=})."
        )
        raise AssertionError(msg)

    with open(filepath) as fh:
        cfgs_raw = json.load(fh)

    cfgs = [
        {
            "run_id": c["paraset_id"],
            **{k.lower(): v for k, v in c["nml_allcfgs"].items()},
        }
        for c in cfgs_raw["configurations"]
    ]

    return cfgs


@define
class AR6SCMRunner:
    """
    Simple climate model runner that follows the same logic as was used in AR6

    If you want exactly the same behaviour as in AR6,
    initialise using [`from_ar6_config`][(c)]
    """

    climate_models_cfgs: dict[str, list[dict[str, Any]]] = field(
        repr=lambda x: ", ".join(
            (
                f"{climate_model}: {len(cfgs)} configurations"
                for climate_model, cfgs in x.items()
            )
        )
    )
    """
    Climate models to run and the configuration to use with them
    """

    output_variables: tuple[str, ...]
    """
    Variables to include in the output
    """

    force_interpolate_to_yearly: bool = True
    """
    Should we interpolate scenarios we run to yearly steps before running the SCMs.
    """

    batch_size_scenarios: int | None = None
    """
    The number of scenarios to run at a time

    Smaller batch sizes use less memory, but take longer overall
    (all else being equal).

    If not supplied, all scenarios are run simultaneously.
    """

    db: OpenSCMDB | None = None
    """
    Database in which to store the output of the runs

    If not supplied, output of the runs is not stored.
    """

    res_column_type: type = int
    """
    Type to cast the result's column type to
    """

    historical_emissions: pd.DataFrame | None = None
    """
    Historical emissions used for harmonisation

    Only required if `run_checks` is `True` to check
    that the data to run is harmonised.
    """

    harmonisation_year: int | None = None
    """
    Year in which the data was harmonised

    Only required if `run_checks` is `True` to check
    that the data to run is harmonised.
    """

    verbose: bool = True
    """
    Should verbose messages be printed?

    This is a temporary hack while we think about how to handle logging
    """

    run_checks: bool = True
    """
    If `True`, run checks on both input and output data

    If you are sure about your workflow,
    you can disable the checks to speed things up
    (but we don't recommend this unless you really
    are confident about what you're doing).
    """

    progress: bool = True
    """
    Should progress bars be shown for each operation?
    """

    n_processes: int | None = multiprocessing.cpu_count()
    """
    Number of processes to use for parallel processing.

    Set to `None` to process in serial.
    """

    def __call__(
        self, in_emissions: pd.DataFrame, force_rerun: bool = False
    ) -> pd.DataFrame:
        """
        Run the simple climate model

        Parameters
        ----------
        in_emissions
            Emissions to run

        force_rerun
            Force scenarios to re-run (i.e. disable caching).

        Returns
        -------
        :
            Raw results from the simple climate model
        """
        if self.run_checks:
            assert_index_is_multiindex(in_emissions)
            assert_has_index_levels(
                in_emissions, ["variable", "unit", "model", "scenario"]
            )
            assert_has_no_pint_incompatible_characters(
                in_emissions.index.get_level_values("unit").unique()
            )
            assert_data_is_all_numeric(in_emissions)

            if self.historical_emissions is None:
                msg = "`self.historical_emissions` must be set to check the infilling"
                raise AssertionError(msg)

            if self.harmonisation_year is None:
                msg = "`self.harmonisation_year` must be set to check the infilling"
                raise AssertionError(msg)

            assert_has_data_for_times(
                in_emissions,
                name="in_emissions",
                times=[self.harmonisation_year, 2100],
                allow_nan=False,
            )

            assert_harmonised(
                in_emissions,
                history=self.historical_emissions,
                harmonisation_time=self.harmonisation_year,
                rounding=5,  # level of data storage in historical data often
            )
            assert_all_groups_are_complete(
                # The combo of the input and infilled should be complete
                in_emissions,
                complete_index=self.historical_emissions.index.droplevel("unit"),
            )

        openscm_runner_emissions = update_index_levels_func(
            in_emissions,
            {
                "variable": partial(
                    convert_variable_name,
                    from_convention=SupportedNamingConventions.GCAGES,
                    to_convention=SupportedNamingConventions.OPENSCM_RUNNER,
                )
            },
        )
        if self.force_interpolate_to_yearly:
            # TODO: put interpolate to annual steps in pandas-openscm
            # Interpolate to ensure no nans.
            for y in range(
                openscm_runner_emissions.columns.min(),
                openscm_runner_emissions.columns.max() + 1,
            ):
                if y not in openscm_runner_emissions:
                    openscm_runner_emissions[y] = np.nan

            openscm_runner_emissions = (
                openscm_runner_emissions.sort_index(axis="columns")
                .T.interpolate("index")
                .T
            )

        scm_results_maybe = run_scms(
            openscm_runner_emissions,
            climate_models_cfgs=self.climate_models_cfgs,
            output_variables=self.output_variables,
            scenario_group_levels=["model", "scenario"],
            n_processes=self.n_processes if self.n_processes is not None else 1,
            db=self.db,
            verbose=self.verbose,
            batch_size_scenarios=self.batch_size_scenarios,
            force_rerun=force_rerun,
        )

        if self.db is not None:
            # Results aren't kept in memory during running, so have to load them now.
            # User can use `run_scms` directly if they want to process differently.
            out_maybe = self.db.load()
            if out_maybe is None:
                raise TypeError(out_maybe)

            out: pd.DataFrame = out_maybe

        else:
            if scm_results_maybe is None:
                raise TypeError(scm_results_maybe)

            out = scm_results_maybe

        out.columns = out.columns.astype(self.res_column_type)

        if self.run_checks:
            # All scenarios have output
            pd.testing.assert_index_equal(  # type: ignore # pandas-stubs out of date
                out.index.droplevel(
                    out.index.names.difference(["model", "scenario"])  # type: ignore # pandas-stubs out of date
                ).drop_duplicates(),
                in_emissions.index.droplevel(
                    in_emissions.index.names.difference(["model", "scenario"])  # type: ignore # pandas-stubs out of date
                ).drop_duplicates(),
                check_order=False,
            )
            # Expected output is provided
            assert_all_groups_are_complete(
                out,
                complete_index=pd.MultiIndex.from_arrays(
                    [list(self.output_variables)], names=["variable"]
                ),
            )

        return out

    @classmethod
    def from_ar6_config(  # noqa: PLR0913
        cls,
        magicc_exe_path: Path,
        magicc_prob_distribution_path: Path,
        output_variables: tuple[str, ...] = DEFAULT_OUTPUT_VARIABLES,
        batch_size_scenarios: int | None = None,
        db: OpenSCMDB | None = None,
        historical_emissions: pd.DataFrame | None = None,
        harmonisation_year: int | None = None,
        verbose: bool = True,
        run_checks: bool = True,
        progress: bool = True,
        n_processes: int | None = multiprocessing.cpu_count(),
    ) -> AR6SCMRunner:
        """
        Initialise from the config used in AR6

        Parameters
        ----------
        magicc_exe_path
            Path to the MAGICC executable to use.

            This should be a MAGICC v7.5.3 executable.

        magicc_prob_distribution_path
            Path to the MAGICC probabilistic distribution.

            This should be the AR6 probabilistic distribution.

        output_variables
            Variables to include in the output

        batch_size_scenarios
            The number of scenarios to run at a time

        db
            Database to use for storing results.

            If not supplied, raw outputs are not stored.

        historical_emissions
            Historical emissions used for harmonisation

            Only required if `run_checks` is `True` to check
            that the data is harmonised before running the SCMs.

        harmonisation_year
            Year in which the data was harmonised

            Only required if `run_checks` is `True` to check
            that the data is harmonised before running the SCMs.

        verbose
            Should verbose messages be printed?

            This is a temporary hack while we think about how to handle logging

        run_checks
            Should checks of the input and output data be performed?

            If this is turned off, things are faster,
            but error messages are much less clear if things go wrong.

        progress
            Should progress bars be shown for each operation?

        n_processes
            Number of processes to use for parallel processing.

            Set to `None` to process in serial.

        Returns
        -------
        :
            Initialised SCM runner
        """
        os.environ["MAGICC_EXECUTABLE_7"] = str(magicc_exe_path)
        check_ar6_magicc7_version()

        magicc_ar6_prob_cfg = load_ar6_magicc_probabilistic_config(
            magicc_prob_distribution_path
        )

        startyear = 1750
        common_cfg = {
            "startyear": startyear,
            "out_dynamic_vars": convert_openscm_runner_output_names_to_magicc_output_names(  # noqa: E501
                output_variables
            ),
            "out_ascii_binary": "BINARY",
            "out_binary_format": 2,
        }

        run_config = [{**common_cfg, **base_cfg} for base_cfg in magicc_ar6_prob_cfg]
        magicc_full_distribution_n_config = 600
        if len(run_config) != magicc_full_distribution_n_config:
            raise AssertionError(len(run_config))

        return cls(
            climate_models_cfgs={"MAGICC7": run_config},
            output_variables=output_variables,
            batch_size_scenarios=batch_size_scenarios,
            db=db,
            historical_emissions=historical_emissions,
            harmonisation_year=harmonisation_year,
            verbose=verbose,
            run_checks=run_checks,
            n_processes=n_processes,
            force_interpolate_to_yearly=True,  # MAGICC safer with annual input
            res_column_type=int,  # annual output by default
        )



"""
Databases used by the package

They're not really databases for the most part
(rather just [pd.DataFrame][pandas.DataFrame]'s),
but the concept is the same.
"""

from __future__ import annotations

from gcages.databases.emissions_variables import EMISSIONS_VARIABLES

__all__ = ["EMISSIONS_VARIABLES"]



"""
Common tools across different approaches
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

import pandas as pd

from gcages.typing import NUMERIC_DATA, TIME_POINT, TimeseriesDataFrame
from gcages.units_helpers import convert_unit_like

if TYPE_CHECKING:
    import pint


class NotHarmonisedError(ValueError):
    """
    Raised when a [pd.DataFrame][pandas.DataFrame] is not harmonised
    """

    def __init__(
        self,
        comparison: pd.DataFrame,
        harmonisation_time: TIME_POINT,
    ) -> None:
        """
        Initialise the error

        Parameters
        ----------
        comparison
            Results of comparing the data and history

        harmonisation_time
            Expected harmonisation time
        """
        error_msg = (
            f"The DataFrame is not harmonised in {harmonisation_time}. "
            f"comparison=\n{comparison}"
        )
        super().__init__(error_msg)


def align_history_to_data_at_time(
    df: TimeseriesDataFrame, *, history: TimeseriesDataFrame, time: Any
) -> tuple[pd.Series[NUMERIC_DATA], pd.Series[NUMERIC_DATA]]:  # type: ignore # pandas-stubs not up to date
    """
    Align history to a given set of data for a given column

    Parameters
    ----------
    df
        Data to which to align history

    history
        History data to align

    time
        Time (i.e. column) for which to align the data

    Returns
    -------
    :
        History, aligned with `df` for the given column

    Raises
    ------
    AssertionError
        `df` and `history` could not be aligned for some reason
    """
    df_year_aligned, history_year_aligned = df[time].align(history[time], join="left")

    # Implicitly assuming that people have already checked
    # that they have history values for all timeseries in `df`,
    # so any null is an obvious issue.
    if history_year_aligned.isnull().any():
        msg_l = ["history did not align properly with df"]

        if df.index.names == history.index.names:
            msg_l.append(
                "history and df have the same index levels "
                f"({list(history.index.names)}). "
                "You probably need to drop some of history's index levels "
                "so alignment can happen along the levels of interest "
                "(usually dropping everything except variable and unit (or similar)). "
            )

        # # Might be useful, pandas might handle it
        # names_only_in_hist = history.index.names.difference(df.index.names)

        for unit_col_guess in ["unit", "units"]:
            if (
                unit_col_guess in df.index.names
                and unit_col_guess in history.index.names
            ):
                df_units_guess = df.index.get_level_values(unit_col_guess)
                history_units_guess = history.index.get_level_values(unit_col_guess)

                differing_units = (
                    df_units_guess.difference(history_units_guess).unique().tolist()
                )
                msg_l.append(
                    "The following units only appear in `df`, "
                    f"which might be why the data isn't aligned: {differing_units}. "
                    f"{df_units_guess=} {history_units_guess=}"
                )

        msg = ". ".join(msg_l)
        raise AssertionError(msg)

    return df_year_aligned, history_year_aligned


def assert_harmonised(  # noqa: PLR0913
    df: TimeseriesDataFrame,
    *,
    history: TimeseriesDataFrame,
    harmonisation_time: TIME_POINT,
    rounding: int = 10,
    df_unit_level: str = "unit",
    history_unit_level: str | None = None,
    ur: pint.UnitRegistry | None = None,
) -> None:
    """
    Assert that the input is harmonised

    Parameters
    ----------
    df
        Data to check

    history
        History to which `df` should be harmonised

    harmonisation_time
        Time at which `df` should be harmonised to `history`

    rounding
        Rounding to apply to the data before comparing

    df_unit_level
        Level in `df`'s index which has unit information

        Only used if unit conversion is required

    history_unit_level
        Level in `history`'s index which has unit information

        If not provided, we assume this is the same as `df_unit_level`

        Only used if unit conversion is required

    ur
        Unit registry to use for determining unit conversions

        Passed to [gcages.units_helpers.convert_unit_like][]

        Only used if unit conversion is required

    Raises
    ------
    NotHarmonisedError
        `df` is not harmonised to `history`
    """
    df_unit_match = convert_unit_like(
        df,
        target=history,
        df_unit_level=df_unit_level,
        target_unit_level=history_unit_level,
        ur=ur,
    )
    df_harm_year_aligned, history_harm_year_aligned = align_history_to_data_at_time(
        df_unit_match, history=history, time=harmonisation_time
    )
    comparison = df_harm_year_aligned.round(rounding).compare(  # type: ignore # pandas-stubs out of date
        history_harm_year_aligned.round(rounding), result_names=("df", "history")
    )
    if not comparison.empty:
        raise NotHarmonisedError(
            comparison=comparison, harmonisation_time=harmonisation_time
        )



"""
General tools for checking completeness
"""

from __future__ import annotations

import pandas as pd


class NotCompleteError(ValueError):
    """
    Raised when a [pd.DataFrame][pandas.DataFrame] is not complete
    """

    def __init__(
        self,
        missing: pd.DataFrame,
        complete_index: pd.MultiIndex,
    ) -> None:
        """
        Initialise the error

        Parameters
        ----------
        missing
            Index levels that don't have a complete index

        complete_index
            Definition of a complete index
        """
        error_msg = (
            "The DataFrame is not complete. "
            f"The following expected levels are missing:\n{missing}\n"
            f"The complete index expected for each level is:\n"
            f"{complete_index.to_frame(index=False)}"
        )
        super().__init__(error_msg)


def get_missing_levels(
    index: pd.MultiIndex,
    complete_index: pd.MultiIndex,
    levels_to_drop: list[str] | None = None,
    unit_col: str | None = None,
) -> pd.MultiIndex:
    """
    Get missing levels in an index

    Here, complete is defined by `complete_index`,
    which specifies the levels that should be in `index`.

    Parameters
    ----------
    index
        Index to check

    complete_index
        Index which defines the meaning of 'complete'

    levels_to_drop
        Levels to drop from `index` before checking for completeness

        If not supplied, we use all the index levels in `index`
        except those that appear in `complete_index` and the `unit_col`,
        specifically
        `index.names.difference([*complete_index.names, unit_col])`.

    unit_col
        Unit column (differences here do not indicate missing levels)

        Only needed if `levels_to_drop` is `None`.

    Examples
    --------
    >>> to_check = pd.MultiIndex.from_tuples(
    ...     [
    ...         ("sa", "va", "W"),
    ...         ("sa", "vb", "W"),
    ...         ("sb", "va", "W"),
    ...         ("sb", "vb", "W"),
    ...     ],
    ...     names=["scenario", "variable", "unit"],
    ... )
    >>> to_check  # doctest: +NORMALIZE_WHITESPACE
    MultiIndex([('sa', 'va', 'W'),
            ('sa', 'vb', 'W'),
            ('sb', 'va', 'W'),
            ('sb', 'vb', 'W')],
           names=['scenario', 'variable', 'unit'])
    >>> # A checker, by which `to_check` is complete
    >>> checker_a = pd.MultiIndex.from_tuples(
    ...     [
    ...         ("va",),
    ...         ("vb",),
    ...     ],
    ...     names=["variable"],
    ... )
    >>> get_missing_levels(to_check, complete_index=checker_a, unit_col="unit")
    MultiIndex([], names=['variable'])
    >>> # Empty index i.e. no missing levels
    >>>
    >>> # A checker which includes variables that aren't present in `to_check`
    >>> checker_b = pd.MultiIndex.from_tuples(
    ...     [
    ...         ("va",),
    ...         ("vb",),
    ...         ("vc",),
    ...     ],
    ...     names=["variable"],
    ... )
    >>> get_missing_levels(
    ...     to_check, complete_index=checker_b, unit_col="unit"
    ... )  # doctest: +NORMALIZE_WHITESPACE
    MultiIndex([('vc',)],
           names=['variable'])
    """
    if levels_to_drop is None:
        if unit_col is None:
            msg = "If levels_to_drop is `None` then `unit_col` must be supplied"
            raise AssertionError(msg)

        # Check against the levels that are in `complete_index`
        # (also ignoring units if they are there)
        levels_to_drop = list(
            {*index.names.difference(complete_index.names), unit_col}  # type: ignore # pandas-stubs confused
        )

    index_to_check = index.droplevel(levels_to_drop)  # type: ignore # something wrong with pandas-stubs

    if not isinstance(index_to_check, pd.MultiIndex):
        index_to_check = pd.MultiIndex.from_arrays(
            [index_to_check.values], names=[index_to_check.name]
        )

    missing_levels: pd.MultiIndex = complete_index.difference(  # type: ignore # pandas-stubs out of date
        index_to_check.reorder_levels(complete_index.names)
    )

    return missing_levels


def assert_all_groups_are_complete(
    to_check: pd.DataFrame,
    complete_index: pd.MultiIndex,
    group_keys: list[str] | None = None,
    unit_col: str = "unit",
) -> None:
    """
    Assert all groups have 'complete' data

    Here, complete is defined by `complete_index`,
    which specifies the metadata that should be included for each group.

    Parameters
    ----------
    to_check
        Data to check

    complete_index
        Index which defines the meaning of 'complete'

    group_keys
        Keys to use to group `to_check` into groups when checking for completeness.

        If not supplied, we use all the index levels in `to_check`
        except those that appear in `complete_index` and the `unit_col`,
        specifically
        `to_check.index.names.difference([*complete_index.names, unit_col])`.

    unit_col
        Unit column (differences here do not indicate incompleteness)

    Raises
    ------
    ValueError
        `to_check` is empty

    NotCompleteError
        `to_check` is not complete

    Examples
    --------
    >>> to_check = pd.DataFrame(
    ...     [
    ...         [1.0, 2.0],
    ...         [3.0, 2.0],
    ...         [1.0, 2.0],
    ...         [3.0, 2.0],
    ...     ],
    ...     columns=[2015, 2100],
    ...     index=pd.MultiIndex.from_tuples(
    ...         [
    ...             ("sa", "va", "W"),
    ...             ("sa", "vb", "W"),
    ...             ("sb", "va", "W"),
    ...             ("sb", "vb", "W"),
    ...         ],
    ...         names=["scenario", "variable", "unit"],
    ...     ),
    ... )
    >>> to_check  # doctest: +NORMALIZE_WHITESPACE
                            2015  2100
    scenario variable unit
    sa       va       W      1.0   2.0
             vb       W      3.0   2.0
    sb       va       W      1.0   2.0
             vb       W      3.0   2.0

    >>> # A checker, by which `to_check` is complete
    >>> checker_a = pd.MultiIndex.from_tuples(
    ...     [
    ...         ("va",),
    ...         ("vb",),
    ...     ],
    ...     names=["variable"],
    ... )
    >>> assert_all_groups_are_complete(to_check, complete_index=checker_a)
    >>> # No error raised, all happy
    >>>
    >>> # A checker which includes variables that aren't present in `to_check`
    >>> checker_b = pd.MultiIndex.from_tuples(
    ...     [
    ...         ("va",),
    ...         ("vb",),
    ...         ("vc",),
    ...     ],
    ...     names=["variable"],
    ... )
    >>> assert_all_groups_are_complete(to_check, complete_index=checker_b)
    Traceback (most recent call last):
        ...
    gcages.completeness.NotCompleteError: The DataFrame is not complete. The following expected levels are missing:
      variable scenario
    0       vc       sa
    0       vc       sb
    The complete index expected for each level is:
      variable
    0       va
    1       vb
    2       vc
    """  # noqa: E501
    if to_check.empty:
        msg = "`to_check` is empty"
        raise ValueError(msg)

    # Probably a smarter way to do this rather than looping, I can't see it now
    if unit_col not in to_check.index.names:
        msg = f"{unit_col=} is not in {to_check.index.names=}"
        raise KeyError(msg)

    if group_keys is None:
        group_keys = to_check.index.names.difference([*complete_index.names, unit_col])  # type: ignore # pandas-stubs confused

    missing_l = []
    # Check against the levels that are in `complete_index`
    # (also ignoring units if they are there)
    idx_to_check_drop_levels = list(
        {*to_check.index.names.difference(complete_index.names), unit_col}  # type: ignore # pandas-stubs confused
    )
    for group_values, gdf in to_check.groupby(group_keys):
        missing_levels = get_missing_levels(
            gdf.index,  # type: ignore # fix when moving to pandas-openscm
            complete_index,
            levels_to_drop=idx_to_check_drop_levels,
        )

        if not missing_levels.empty:
            tmp = missing_levels.to_frame(index=False)
            # Could probably do this better too if we need speed
            for key, value in zip(group_keys, group_values):
                tmp[key] = value

            missing_l.append(tmp)

    if missing_l:
        raise NotCompleteError(pd.concat(missing_l), complete_index=complete_index)



"""
General simple climate model (SCM) running tools
"""

from __future__ import annotations

import os
from collections.abc import Iterable
from typing import Any, Optional, cast

import pandas as pd
from pandas_openscm.db import EmptyDBError, OpenSCMDB
from pandas_openscm.indexing import multi_index_lookup, multi_index_match
from pandas_openscm.parallelisation import ParallelOpConfig

from gcages.exceptions import MissingOptionalDependencyError


def convert_openscm_runner_output_names_to_magicc_output_names(
    openscm_runner_names: Iterable[str],
) -> tuple[str, ...]:
    """
    Get output names for the call to MAGICC

    Parameters
    ----------
    openscm_runner_names
        OpenSCM-Runner output names

    Returns
    -------
    :
        MAGICC output names
    """
    # TODO: move this to OpenSCM-Runner or fix up pymagicc
    # (not doing now because of the headache of upgrading those packages)
    try:
        import pymagicc.definitions  # type: ignore
    except ImportError as exc:
        raise MissingOptionalDependencyError(
            "convert_openscm_runner_output_names_to_magicc_output_names",
            requirement="pymagicc",
        ) from exc

    res_l = []
    for openscm_runner_variable in openscm_runner_names:
        if openscm_runner_variable == "Surface Air Temperature Change":
            # A fun inconsistency
            res_l.append("SURFACE_TEMP")
        elif openscm_runner_variable == "Effective Radiative Forcing|HFC4310mee":
            # Another fun inconsistency
            magicc_var = pymagicc.definitions.convert_magicc7_to_openscm_variables(
                "Effective Radiative Forcing|HFC4310",
                inverse=True,
            )
            res_l.append(magicc_var)
        else:
            magicc_var = pymagicc.definitions.convert_magicc7_to_openscm_variables(
                openscm_runner_variable,
                inverse=True,
            )
            res_l.append(magicc_var)

    return tuple(res_l)


def batch_df(  # noqa: D103
    df: pd.DataFrame, batch_index: pd.MultiIndex, batch_size: int | None
) -> list[pd.DataFrame]:
    # TOOD: move this to pandas-openscm
    if batch_size is None:
        batches = [df]

    else:
        batches = []
        for i in range(0, batch_index.size, batch_size):
            start = i
            stop = min(i + batch_size, batch_index.shape[0])

            batches.append(multi_index_lookup(df, batch_index[start:stop]))

    return batches


def run_batch(
    batch: pd.DataFrame,
    climate_models_cfgs: dict[str, list[dict[str, Any]]],
    output_variables: tuple[str, ...],
) -> pd.DataFrame:
    """
    Run a batch of scenarios

    Parameters
    ----------
    batch
        Batch to run

    climate_models_cfgs
        Climate model to run and its configuration

        Passed to [openscm_runner.run.run]

    output_variables
        Output variables to retrieve from the climate model

        Passed to [openscm_runner.run.run]

    Returns
    -------
    :
        Results of running the batch
    """
    # TODO: move to openscm-runner
    # (not there at the moment because of maintenance issues)

    try:
        import openscm_runner.run
    except ImportError as exc:
        raise MissingOptionalDependencyError(
            "run_batch", requirement="openscm_runner"
        ) from exc

    try:
        import scmdata
    except ImportError as exc:
        raise MissingOptionalDependencyError(
            "run_batch", requirement="scmdata"
        ) from exc

    batch_res = openscm_runner.run.run(  # type: ignore
        scenarios=scmdata.ScmRun(batch, copy_data=True),
        climate_models_cfgs=climate_models_cfgs,
        output_variables=output_variables,
    ).timeseries(time_axis="year")

    return cast(pd.DataFrame, batch_res)


def get_scenarios_to_run_after_checking_cache(  # noqa: PLR0913
    scenarios: pd.DataFrame,
    climate_model: str,
    db: OpenSCMDB,
    scenario_group_levels: list[str],
    climate_model_level: str,
    verbose: bool,
) -> pd.DataFrame | None:
    """
    Get the scenarios to run after checking the database cache

    Parameters
    ----------
    scenarios
        Full set of scenarios

    climate_model
        Climate model we are going to run with

    db
        Database in which results are being stored

    scenario_group_levels
        Index levels which define scenario groups

        Typically something like ["model", "scenario"]

    climate_model_level
        Climate model level in the database's metadata

        This level should store information
        about the climate model used to run the scenario.

    verbose
        If we skip running scenarios, should we print a message showing which ones

    Returns
    -------
    :
        Scenarios to run

        If all scenarios have already been run for `climate_model`,
        we return `None`
    """
    # TODO: move to openscm-runner
    # (not there at the moment because of maintenance issues)
    try:
        existing_metadata = db.load_metadata()
    except EmptyDBError:
        # Empty DB, we know we need to run everything
        return scenarios

    check_levels = [*scenario_group_levels, climate_model_level]
    db_already_run = existing_metadata.droplevel(
        existing_metadata.names.difference(check_levels)  # type: ignore # pandas-stubs out of date
    ).unique()

    # TODO: set_new_single_value_levels into pandas-openscm's index_manipulation
    new_values = [climate_model]
    new_names = [climate_model_level]
    if not isinstance(scenarios.index, pd.MultiIndex):
        raise TypeError(scenarios.index)

    batch_output_exp_index = pd.MultiIndex(
        codes=[
            *scenarios.index.codes,
            *([[0] * scenarios.index.shape[0]] * len(new_values)),
        ],
        levels=[*scenarios.index.levels, *[pd.Index([value]) for value in new_values]],
        names=[*scenarios.index.names, *new_names],
    )

    already_run_idx = multi_index_match(batch_output_exp_index, db_already_run)
    batch_to_run = scenarios.loc[~already_run_idx, :]
    already_run = scenarios.loc[already_run_idx, :]
    already_run_disp = already_run.index.droplevel(
        already_run.index.names.difference(check_levels)  # type: ignore # pandas-stubs out of date
    ).unique()
    if not already_run_disp.empty and verbose:
        # There are nicer ways to do this than verbose,
        # but thinking through logging is a problem for another day
        # (making loguru a required dependency might be the answer,
        # I don't know if it has any other dependencies).
        print(
            "Not re-running already run scenarios:\n"
            f"{already_run_disp.to_frame(index=False)}"
        )

    if batch_to_run.empty:
        return None

    return batch_to_run


def run_scms(  # noqa: PLR0912, PLR0913
    scenarios: pd.DataFrame,
    climate_models_cfgs: dict[str, list[dict[str, Any]]],
    output_variables: tuple[str, ...],
    scenario_group_levels: list[str],
    n_processes: int,
    db: OpenSCMDB | None = None,
    db_climate_model_level: str = "climate_model",
    verbose: bool = True,
    progress: bool = True,
    batch_size_scenarios: int | None = None,
    force_rerun: bool = False,
) -> pd.DataFrame | None:
    """
    Run simple climate models (SCMs)

    Parameters
    ----------
    scenarios
        Scenarios to run

    climate_models_cfgs
        Climate model to run and its configuration

        Passed to [openscm_runner.run.run]

    output_variables
        Output variables to retrieve from the climate model

        Passed to [openscm_runner.run.run]

    scenario_group_levels
        Index levels which define scenario groups

        Typically something like ["model", "scenario"]

    n_processes
        Number of parallel processes to use while running

    db
        Database in which to save the results

        If not provided, results are not saved along the way

    db_climate_model_level
        Climate model level in the database's metadata

        This level should store information
        about the climate model used to run the scenario.

    verbose
        If we skip running scenarios because they have already been run,
        should we print a message showing which ones?

    progress
        Should progress bar(s) be displayed?

    batch_size_scenarios
        How many scenarios should be run in a single batch?

        Running more scenarios at once is faster,
        but it runs the risk of running out of memory.

    force_rerun
        Should we force the scenarios to be re-run, even if they are already in `db`

    Returns
    -------
    :
        Results of running the SCM

        If `db` is provided, returns `None`
        (so you as the user can decided whether all the output
        should be in memory at once)
    """
    # TODO: move to openscm-runner
    # (not there at the moment because of maintenance issues)
    try:
        import openscm_runner.adapters
    except ImportError as exc:
        raise MissingOptionalDependencyError(
            "run_scms", requirement="openscm_runner"
        ) from exc

    scens_to_run = scenarios.index.droplevel(
        scenarios.index.names.difference(scenario_group_levels)  # type: ignore # pandas-stubs out of date
    ).unique()
    climate_models_cfgs_iter = climate_models_cfgs.items()
    if progress:
        pconfig = ParallelOpConfig.from_user_facing(
            progress=progress,
            progress_results_kwargs=dict(desc="Climate models"),
            max_workers=None,  # This loop always goes in serial
        )
        climate_models_cfgs_iter = pconfig.progress_results(  # type: ignore # something weird happening here
            climate_models_cfgs_iter, desc="Climate models"
        )

    for climate_model, cfg in climate_models_cfgs_iter:
        cfg_use = cfg
        if force_rerun or db is None:
            scenarios_use: Optional[pd.DataFrame] = scenarios

        else:
            if climate_model == "MAGICC7":
                # Urgh
                climate_model_check = (
                    f"MAGICC{openscm_runner.adapters.MAGICC7.get_version()}"  # type: ignore
                )
            else:
                climate_model_check = climate_model

            scenarios_use = get_scenarios_to_run_after_checking_cache(
                scenarios,
                climate_model=climate_model_check,
                db=db,
                scenario_group_levels=scenario_group_levels,
                climate_model_level=db_climate_model_level,
                verbose=verbose,
            )
            if scenarios_use is None:
                # Already all run for this climate model
                continue

        if climate_model == "MAGICC7":
            # Avoid MAGICC's last year jump
            magicc_extra_years = 3
            cfg_use = [
                {**c, "endyear": scenarios.columns.max() + magicc_extra_years}
                for c in cfg_use
            ]
            os.environ["MAGICC_WORKER_NUMBER"] = str(n_processes)

            if scenarios_use is None:
                raise TypeError(scenarios_use)
            scenarios_use = scenarios_use.copy()
            last_year = scenarios_use.columns.max()
            scenarios_use[last_year + magicc_extra_years] = scenarios_use[last_year]
            scenarios_use = (
                scenarios_use.sort_index(axis="columns").T.interpolate("index").T
            )

        if scenarios_use is None:
            raise TypeError(scenarios_use)

        scenario_batches = batch_df(
            scenarios_use,
            batch_index=scens_to_run,
            batch_size=batch_size_scenarios,
        )

        if progress:
            pconfig = ParallelOpConfig.from_user_facing(
                progress=progress,
                progress_results_kwargs=dict(desc="Scenario batches"),
                max_workers=None,
            )
            scenario_batches = pconfig.progress_results(  # type: ignore # not sure what is happening here
                scenario_batches, desc="Scenario batch"
            )

        if db is None:
            res_l = []

        for scenario_batch in scenario_batches:
            batch_res = run_batch(
                batch=scenario_batch,
                climate_models_cfgs={climate_model: cfg_use},
                output_variables=output_variables,
            )

            if climate_model == "MAGICC7":
                # Chop off the extra years
                batch_res = batch_res.iloc[:, :-magicc_extra_years]
                # Chop out regional results
                batch_res = batch_res.loc[
                    batch_res.index.get_level_values("region") == "World"
                ]

            if db is not None:
                db.save(batch_res)
            else:
                res_l.append(batch_res)

    if db is not None:
        # Assume the user doesn't want all the data in memory at once
        # (I really don't like this pattern, but can't fix it now)
        return None

    res = pd.concat(res_l)

    return res



"""
Exceptions that are used throughout
"""

from __future__ import annotations

import difflib
from collections.abc import Collection
from typing import Any


class MissingOptionalDependencyError(ImportError):
    """
    Raised when an optional dependency is missing

    For example, plotting dependencies like matplotlib
    """

    def __init__(self, callable_name: str, requirement: str) -> None:
        """
        Initialise the error

        Parameters
        ----------
        callable_name
            The name of the callable that requires the dependency

        requirement
            The name of the requirement
        """
        error_msg = f"`{callable_name}` requires {requirement} to be installed"
        super().__init__(error_msg)


class UnrecognisedValueError(ValueError):
    """
    Raised when a value is not recognised

    In this context, recognised means 'known'
    in the sense of being part of some set of values
    that are understood and defined.
    For example, this error could be raised when a value
    is not be part of a set of a controlled vocabulary/known definitions.
    """

    def __init__(
        self, unrecognised_value: Any, name: Any, known_values: Collection[Any]
    ) -> None:
        """
        Initialise the error

        Parameters
        ----------
        unrecognised_value
            The unrecognised value

        name
            The name of the thing that has the unrecognised value

            For example, `unrecognised_value` might be `"Emissions|junk"`
            and `name` could be `"variable"`.

        known_values
            The known values for `metadata_key`
        """
        error_msg_l = [f"{unrecognised_value!r} is not a recognised value for {name}."]
        close = difflib.get_close_matches(
            unrecognised_value, known_values, n=3, cutoff=0.6
        )
        if close:
            close_with_quotes = [f"{v!r}" for v in close]
            error_msg_l.append(f"Did you mean {' or '.join(close_with_quotes)}?")

        error_msg_l.append(f"The full list of known values is: {known_values}")

        super().__init__(" ".join(error_msg_l))



"""
Renaming between naming conventions

At present, the naming convention for both sides is implicit.
We are considering adding the concept of controlled vocabularies
to help clarify this, but have not done so yet.
"""

from __future__ import annotations

import sys
from typing import cast

import pandas as pd

import gcages.databases
from gcages.exceptions import UnrecognisedValueError

if sys.version_info >= (3, 11):
    from enum import StrEnum
else:
    from backports.strenum import StrEnum


class SupportedNamingConventions(StrEnum):
    """Supported naming conventions"""

    GCAGES = "gcages"
    """This package's naming convention"""

    AR6_CFC_INFILLING_DB = "ar6_cfc_infilling_db"
    """
    The naming convention used in AR6's CFC infilling database

    Somehow this ended up being different to all the other conventions.
    """

    IAMC = "iamc"
    """
    Integrated Assessment Modelling Consortium (IAMC) naming convention

    Not a perfect definition so the implementation here is a bit of a guess
    based on experience.
    https://github.com/IAMconsortium/common-definitions
    is a better source of truth, but it also moves more quickly,
    is not used universally and covers many more variables
    than we care about within the gcages context.
    """

    CMIP7_SCENARIOMIP = "cmip7_scenariomip"
    """
    The naming convention used during preparation of ScenarioMIP for CMIP7
    """

    OPENSCM_RUNNER = "openscm_runner"
    """
    OpenSCM-Runner naminv convention

    Used by the package which actually runs simple climate models (SCMs),
    see https://github.com/openscm/openscm-runner
    """

    RCMIP = "rcmip"
    """
    Reduced Complexity Model Intercomparison Project (RCMIP) naming convention

    See rcmip.org and e.g.

    - https://rcmip-protocols-au.s3-ap-southeast-2.amazonaws.com/v5.1.0/rcmip-emissions-annual-means-v5-1-0.csv
    - https://rcmip-protocols-au.s3-ap-southeast-2.amazonaws.com/v5.1.0/rcmip-concentrations-annual-means-v5-1-0.csv
    - https://rcmip-protocols-au.s3-ap-southeast-2.amazonaws.com/v5.1.0/rcmip-radiative-forcing-annual-means-v5-1-0.csv
    """


def convert_variable_name(
    variable_in: str,
    from_convention: SupportedNamingConventions,
    to_convention: SupportedNamingConventions,
    database: pd.DataFrame = gcages.databases.EMISSIONS_VARIABLES,
) -> str:
    """
    Convert a variable name to another naming convention

    Parameters
    ----------
    variable_in
        Variable name to convert

    from_convention
        Convention to convert from

    to_convention
        Convention to convert to

    database
        Database to use for the conversions

    Returns
    -------
    :
        Converted variable name

    Raises
    ------
    UnrecognisedValueError
        `variable_in` is not a recognised value in `from_convention`
    """
    from_key = str(from_convention)
    to_key = str(to_convention)

    res_l = database.loc[database[from_key] == variable_in, to_key].tolist()

    if len(res_l) < 1:
        raise UnrecognisedValueError(
            unrecognised_value=variable_in,
            name=f"the {from_convention} naming convention",
            known_values=sorted(list(database[from_key].tolist())),
        )

    if len(res_l) > 1:  # pragma: no cover
        raise AssertionError(res_l)

    return cast(str, res_l[0])



"""
Type hints that are used throughout
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Union

import numpy as np
import numpy.typing as npt
import pandas as pd

if TYPE_CHECKING:
    import pint
    from typing_extensions import TypeAlias

NP_FLOAT_OR_INT: TypeAlias = Union[np.floating[Any], np.integer[Any]]
"""
Type alias for a numpy float or int (not complex)
"""

NP_ARRAY_OF_FLOAT_OR_INT: TypeAlias = npt.NDArray[NP_FLOAT_OR_INT]
"""
Type alias for an array of numpy float or int (not complex)
"""

NUMERIC_DATA: TypeAlias = Union[float, int, NP_FLOAT_OR_INT]
"""
Type alias for a value that can be used in the data of a [TimeseriesDataFrame][(m).]
"""

TIME_POINT: TypeAlias = Union[float, int]
"""
Type alias for a value that can be used in the columns of a [TimeseriesDataFrame][(m).]
"""

TimeseriesDataFrame: TypeAlias = pd.DataFrame
"""
Type alias for the [pd.DataFrame][pandas.DataFrame] shape we use throughout

For typing purposes, this is just a direct alias of [pd.DataFrame][pandas.DataFrame].
However, the point of defining this
is to provide greater clarity of the kind of data we expect.

We expect a collection of timeseries.
These timeseries are defined by the columns, which we expect to be timepoints.
We expect that the index contains metadata about each timeseries.
As a result, the data itself should be numerical only (no strings, no lists, no dicts).

An example of this kind of data is given below.
Note, in line with the description above:

1. Data is in the body of the [pd.DataFrame][pandas.DataFrame]
1. The columns define the time axis
1. All metadata is contained in the index

```python
                        2015  2100
scenario variable unit
sa       va       W      1.0   2.1
         vb       W      3.0   2.0
```
"""

if TYPE_CHECKING:
    PINT_SCALAR: TypeAlias = pint.facets.numpy.quantity.NumpyQuantity[NP_FLOAT_OR_INT]
    """
    Type alias for a pint quantity that wraps a numpy scalar
    """



"""General harmonisation tools"""

from __future__ import annotations

from gcages.harmonisation.aneris import AnerisHarmoniser
from gcages.harmonisation.common import NotHarmonisedError, assert_harmonised

__all__ = ["AnerisHarmoniser", "NotHarmonisedError", "assert_harmonised"]



"""
Infilling part of the AR6 workflow
"""

from __future__ import annotations

import functools
import platform
from collections.abc import Iterable, Mapping
from functools import partial
from pathlib import Path
from typing import TYPE_CHECKING, Callable, cast

import pandas as pd
from attrs import define
from pandas_openscm.index_manipulation import update_index_levels_func
from pandas_openscm.io import load_timeseries_csv
from pandas_openscm.parallelisation import ParallelOpConfig, apply_op_parallel_progress

from gcages.assertions import (
    assert_data_is_all_numeric,
    assert_has_index_levels,
    assert_index_is_multiindex,
    assert_only_working_on_variable_unit_variations,
)
from gcages.completeness import assert_all_groups_are_complete
from gcages.exceptions import MissingOptionalDependencyError
from gcages.harmonisation import assert_harmonised
from gcages.hashing import get_file_hash
from gcages.renaming import SupportedNamingConventions, convert_variable_name

if TYPE_CHECKING:
    import silicone  # type: ignore


@functools.cache
def load_massaged_ar6_infilling_db(filepath: Path, cfcs: bool) -> pd.DataFrame:
    """
    Load the infilling database that was used in AR6

    The data is massaged to what is expected by our infilling,
    it isn't the raw data (at least not raw metadata).

    Parameters
    ----------
    filepath
        Filepath from which to load the infilling database

    cfcs
        If `True`, we expect to load the database for infilling CFCs

    Returns
    -------
    :
        Infilling database used in AR6

    Raises
    ------
    AssertionError
        `filepath` points to a file that does not have the expected hash
    """
    fp_hash = get_file_hash(filepath, algorithm="sha256")
    if cfcs:
        if platform.system() == "Windows":
            fp_hash_exp = (
                "39ca03de170dab52a1845a25ebedeeb69704b8c86d16be14a08cc06fceba5369"
            )
        else:
            fp_hash_exp = (
                "9ad0550c671701622ec2b2e7ba2b6c38d58f83507938ab0f5aa8b1a35d26c015"
            )

    else:
        fp_hash_exp = "4ef7aabb18c35fdf857145adfbfffbbcfd6667b5551cc0b7a182e682f03b5843"

    if fp_hash != fp_hash_exp:
        msg = (
            f"The sha256 hash of {filepath} is {fp_hash}. "
            f"This does not match what we expect ({fp_hash_exp=})."
        )
        raise AssertionError(msg)

    res = load_timeseries_csv(
        filepath,
        lower_column_names=True,
        index_columns=["model", "scenario", "variable", "region", "unit"],
        out_columns_type=int,
    )

    if cfcs:
        # Not sure why this was like this, anyway
        res = update_index_levels_func(
            res,
            {
                # Another naming convention?
                # Probably have to check RCMIP...
                "variable": lambda x: x.replace("|HFC|", "|")
                .replace("|PFC|", "|")
                .replace("Energy and Industrial Processes", "Fossil"),
            },
        )

    else:
        res = update_index_levels_func(
            res,
            {
                "variable": lambda x: x.replace(
                    "AR6 climate diagnostics|Harmonized|", ""
                )
            },
        )
        res = res.loc[
            ~res.index.get_level_values("variable").isin(
                # Variables that aren't actually used
                ["Emissions|F-Gases", "Emissions|HFC", "Emissions|PFC"]
            )
        ]
        res = update_index_levels_func(
            res,
            {
                "variable": partial(
                    convert_variable_name,
                    from_convention=SupportedNamingConventions.IAMC,
                    to_convention=SupportedNamingConventions.GCAGES,
                )
            },
        )

    return res


@functools.cache
def get_ar6_full_historical_emissions(filepath: Path) -> pd.DataFrame:
    """
    Get the full AR6 historical emissions

    Parameters
    ----------
    filepath
        Filepath from which to load the emissions

    Returns
    -------
    :
        Historical emissions as used in AR6

    Raises
    ------
    AssertionError
        `filepath` points to a file that does not have the expected hash
    """
    try:
        from pandas_indexing.core import assignlevel, projectlevel
        from pandas_indexing.selectors import isin, ismatch
    except ImportError as exc:
        raise MissingOptionalDependencyError(
            "get_ar6_full_historical_emissions", requirement="pandas_indexing"
        ) from exc

    fp_hash = get_file_hash(filepath, algorithm="sha256")
    if platform.system() == "Windows":
        fp_hash_exp = "39ca03de170dab52a1845a25ebedeeb69704b8c86d16be14a08cc06fceba5369"
    else:
        fp_hash_exp = "9ad0550c671701622ec2b2e7ba2b6c38d58f83507938ab0f5aa8b1a35d26c015"

    if fp_hash != fp_hash_exp:
        msg = (
            f"The sha256 hash of {filepath} is {fp_hash}. "
            f"This does not match what we expect ({fp_hash_exp=})."
        )
        raise AssertionError(msg)

    raw = load_timeseries_csv(
        filepath,
        lower_column_names=True,
        index_columns=["model", "scenario", "variable", "region", "unit"],
        out_columns_type=int,
    )

    history = raw.loc[
        isin(scenario="ssp245") & ~ismatch(variable="**CO2"), :2015
    ].reset_index(["model", "scenario"], drop=True)

    history = assignlevel(
        history,
        variable=projectlevel(history.index, "variable").map(
            lambda x: convert_variable_name(
                x,
                from_convention=SupportedNamingConventions.AR6_CFC_INFILLING_DB,
                to_convention=SupportedNamingConventions.GCAGES,
            )
        ),
        # Not strictly necessary, but makes life easier
        unit=projectlevel(history.index, "unit").map(
            lambda x: x.replace("HFC4310mee/yr", "HFC4310/yr").replace(
                "NO2 / yr", "NO2/yr"
            )
        ),
    )
    # Not sure why this happened, but here we are
    history.loc[ismatch(variable="**HFC245fa"), :] *= 0.0  # type: ignore # pix typing not playing nice with pandas-stubs

    return history


def infill_scenario(
    indf: pd.DataFrame,
    infillers: Mapping[str, Callable[[pd.DataFrame], pd.DataFrame]],
) -> pd.DataFrame:
    """
    Infill a single scenario

    Parameters
    ----------
    indf
        Scenario to harmonise

    infillers
        Functions to use for infilling each variable.

        The keys define the variable that can be infilled.
        The variables define the function which,
        given inputs with the expected lead variables,
        returns the infilled time series.

    Returns
    -------
    :
        Infilled scenario
    """
    try:
        from pandas_indexing.core import concat, uniquelevel
        from pandas_indexing.selectors import isin
    except ImportError as exc:
        raise MissingOptionalDependencyError(
            "infill_scenario", requirement="pandas_indexing"
        ) from exc

    assert_only_working_on_variable_unit_variations(indf)

    indf_variables = uniquelevel(indf, "variable")

    inferred_l = []
    # Don't repeat this logic again, it is super confusing
    # and should be done in pre-processing instead.
    if "Emissions|CO2" in indf_variables:
        for component_missing, component_other in (
            ("Emissions|CO2|Fossil", "Emissions|CO2|Biosphere"),
            ("Emissions|CO2|Biosphere", "Emissions|CO2|Fossil"),
        ):
            if (
                component_missing not in indf_variables
                and component_other in indf_variables
            ):
                inferred = (
                    indf.loc[isin(variable=["Emissions|CO2"])]
                    .subtract(
                        indf.loc[isin(variable=[component_other])].reset_index(
                            "variable", drop=True
                        ),
                        axis="rows",  # type: ignore # pandas-stubs confused
                    )
                    .pix.assign(variable=component_missing)
                )
                inferred_l.append(inferred)

    if inferred_l:
        inferred = concat(inferred_l)
        inferred_variables = inferred.pix.unique("variable")
    else:
        inferred = None
        inferred_variables = []

    to_infill_silicone = [
        v for v in infillers if v not in indf_variables and v not in inferred_variables
    ]

    infilled_silicone_l = []
    for v_to_infill in to_infill_silicone:
        infiller = infillers[v_to_infill]
        tmp = infiller(indf)
        # The fact that this is needed suggests there's a bug in silicone
        tmp = tmp.loc[:, indf.columns]  # type: ignore # pandas-stubs being stupid
        infilled_silicone_l.append(tmp)

    # Also add zeros for HFC245fa.
    # The fact that this is effectively hidden is an illustration
    # of the issues with how our stack is set up.
    tmp = (tmp * 0.0).pix.assign(variable="Emissions|HFC245fa", unit="kt HFC245fa/yr")
    infilled_silicone_l.append(tmp)

    infilled_silicone = concat(infilled_silicone_l).sort_index(axis="columns")
    infilled_silicone_vars = uniquelevel(infilled_silicone, "variable")

    # If we started with total CO2 then infilled fossil CO2,
    # we preserve harmonisation and the total by overwriting AFOLU CO2.
    if (
        "Emissions|CO2" in indf_variables
        and "Emissions|CO2|Fossil" in infilled_silicone_vars
    ):
        infilled_silicone = concat(
            [
                infilled_silicone.loc[~isin(variable=["Emissions|CO2|Biosphere"])],
                (
                    indf.loc[isin(variable=["Emissions|CO2"])]
                    .subtract(
                        infilled_silicone.loc[
                            isin(variable=["Emissions|CO2|Fossil"])
                        ].reset_index("variable", drop=True),
                        axis="rows",  # type: ignore # pandas-stubs confused
                    )
                    .pix.assign(variable="Emissions|CO2|Biosphere")
                ),
            ]
        )

    if inferred is not None:
        infilled: pd.DataFrame = concat([inferred, infilled_silicone])
    else:
        infilled = infilled_silicone

    return infilled


@functools.cache
def get_ar6_infiller(  # type: ignore # silicone has no type hints
    follower: str,
    lead: tuple[str, ...],
    db_file: Path,
    cruncher: silicone.base._DatabaseCruncher,
    cfcs: bool,
) -> Callable[[pd.DataFrame], pd.DataFrame]:
    """
    Get an infiller following the logic used in AR6

    Parameters
    ----------
    follower
        Variable to infill

    lead
        Lead variable(s) for the infilling

    db_file
        File from which to load the infilling database

    cruncher
        Cruncher to use for deriving the relationship between `follower` and `lead`

    cfcs
        Should we load the CFC infilling database?

    Returns
    -------
    :
        Infilled timeseries
    """
    try:
        import pyam  # type: ignore # pyam not typed
    except ImportError as exc:
        raise MissingOptionalDependencyError(
            "get_ar6_infiller", requirement="pyam"
        ) from exc

    db = load_massaged_ar6_infilling_db(db_file, cfcs=cfcs)

    infiller = cruncher(pyam.IamDataFrame(db)).derive_relationship(
        variable_follower=follower,
        variable_leaders=list(lead),
    )

    def res_func(inp: pd.DataFrame) -> pd.DataFrame:
        return cast(pd.DataFrame, infiller(pyam.IamDataFrame(inp)).timeseries())

    return res_func


def do_ar6_like_infilling(  # type: ignore # noqa: PLR0913 # silicone has no type hints
    indf: pd.DataFrame,
    follower: str,
    lead_options: tuple[tuple[str, ...], ...],
    db_file: Path,
    cruncher: silicone.base._DatabaseCruncher,
    cfcs: bool,
) -> pd.DataFrame:
    """
    Do infilling like it was done in AR6

    Parameters
    ----------
    indf
        Emissions scenario to infill

    follower
        Variable to infill

    lead_options
        Options to try as the lead variable

    db_file
        File from which to load the infilling database

    cruncher
        Cruncher to use for deriving the relationship between `follower` and `lead`

    cfcs
        Should we load the CFC infilling database?

    Returns
    -------
    :
        Infilled timeseries
    """
    indf_variables = indf.index.get_level_values("variable")
    for leads in lead_options:
        if all(v in indf_variables for v in leads):
            # This function is split out to allow for caching
            infiller = get_ar6_infiller(
                follower=follower,
                lead=leads,
                db_file=db_file,
                cruncher=cruncher,
                cfcs=cfcs,
            )

            res = infiller(indf)
            break

    else:
        raise NotImplementedError

    return res


@define
class AR6Infiller:
    """
    Infiller that follows the same logic as was used in AR6

    If you want exactly the same behaviour as in AR6,
    initialise using [`from_ar6_config`][(c)]
    """

    infillers: Mapping[str, Callable[[pd.DataFrame], pd.DataFrame]]
    """
    Functions to use for infilling each variable.

    The keys define the variable that can be infilled.
    The variables define the function which,
    given inputs with the expected lead variables,
    returns the infilled timeseries.
    """

    run_checks: bool = True
    """
    If `True`, run checks on both input and output data

    If you are sure about your workflow,
    you can disable the checks to speed things up
    (but we don't recommend this unless you really
    are confident about what you're doing).
    """

    historical_emissions: pd.DataFrame | None = None
    """
    Historical emissions used for harmonisation

    Only required if `run_checks` is `True` to check
    that the infilled data is also harmonised.
    """

    harmonisation_year: int | None = None
    """
    Year in which the data was harmonised

    Only required if `run_checks` is `True` to check
    that the infilled data is also harmonised.
    """

    progress: bool = True
    """
    Should progress bars be shown for each operation?
    """

    n_processes: int | None = None  # better off in serial with silicone
    """
    Number of processes to use for parallel processing.

    Set to `None` to process in serial.
    """

    def __call__(self, in_emissions: pd.DataFrame) -> pd.DataFrame:
        """
        Infill

        Parameters
        ----------
        in_emissions
            Emissions to infill

        Returns
        -------
        :
            Infilled emissions
        """
        if self.run_checks:
            assert_index_is_multiindex(in_emissions)
            assert_data_is_all_numeric(in_emissions)
            # Needed for parallelisation
            assert_has_index_levels(
                in_emissions, ["variable", "unit", "model", "scenario"]
            )

        # Strip off any prefixes that might be there
        to_infill = in_emissions.pix.assign(
            variable=in_emissions.index.get_level_values("variable").map(
                lambda x: x.replace("AR6 climate diagnostics|Harmonized|", "")
            )
        ).sort_index(axis="columns")

        # This is not the most efficient parallelisation.
        # It would be better to group by variable,
        # then have an `infill_variable` function,
        # but this would require a bit more thinking
        # to actually achieve.
        infilled = pd.concat(
            [
                v.reorder_levels(to_infill.index.names)
                for v in apply_op_parallel_progress(
                    func_to_call=infill_scenario,
                    iterable_input=(
                        gdf for _, gdf in to_infill.groupby(["model", "scenario"])
                    ),
                    parallel_op_config=ParallelOpConfig.from_user_facing(
                        progress=self.progress,
                        max_workers=self.n_processes,
                        progress_results_kwargs=dict(desc="Scenarios to infill"),
                    ),
                    infillers=self.infillers,
                )
            ]
        )

        if self.run_checks:
            pd.testing.assert_index_equal(infilled.columns, in_emissions.columns)

            if self.historical_emissions is None:
                msg = "`self.historical_emissions` must be set to check the infilling"
                raise AssertionError(msg)

            if self.harmonisation_year is None:
                msg = "`self.harmonisation_year` must be set to check the infilling"
                raise AssertionError(msg)

            assert_harmonised(
                infilled,
                history=self.historical_emissions,
                harmonisation_time=self.harmonisation_year,
                rounding=5,  # level of data storage in historical data often
            )
            assert_all_groups_are_complete(
                # The combo of the input and infilled should be complete
                pd.concat(
                    [in_emissions, infilled.reorder_levels(in_emissions.index.names)]
                ),
                complete_index=self.historical_emissions.index.droplevel("unit"),
            )

        return infilled

    @classmethod
    def from_ar6_config(  # noqa: PLR0913
        cls,
        ar6_infilling_db_file: Path,
        ar6_infilling_db_cfcs_file: Path,
        variables_to_infill: Iterable[str] | None = None,
        run_checks: bool = True,
        historical_emissions: pd.DataFrame | None = None,
        harmonisation_year: int | None = None,
        progress: bool = True,
        n_processes: int | None = None,  # better off in serial with silicone
    ) -> AR6Infiller:
        """
        Initialise from the config used in AR6

        Parameters
        ----------
        ar6_infilling_db_file
            File containing the AR6 infilling database

            This is for all emissions except CFCs.

        ar6_infilling_db_cfcs_file
            File containing the AR6 infilling database for CFCs

        variables_to_infill
            Variables to infill.

            If not supplied, we use the default set from AR6.

        run_checks
            Should checks of the input and output data be performed?

            If this is turned off, things are faster,
            but error messages are much less clear if things go wrong.

        historical_emissions
            Historical emissions used for harmonisation

            Only required if `run_checks` is `True` to check
            that the infilled data is also harmonised.

        harmonisation_year
            Year in which the data was harmonised

            Only required if `run_checks` is `True` to check
            that the infilled data is also harmonised.

        progress
            Should a progress bar be shown for each operation?

        n_processes
            Number of processes to use for parallel processing.

            Set to `None` to process in serial.

        Returns
        -------
        :
            Initialised harmoniser
        """
        try:
            import silicone.database_crunchers  # type: ignore # silicone has no type hints
        except ImportError as exc:
            raise MissingOptionalDependencyError(
                "get_ar6_infiller", requirement="silicone"
            ) from exc

        VARS_DB_CRUNCHERS = {
            "Emissions|BC": (
                False,
                silicone.database_crunchers.QuantileRollingWindows,
            ),
            "Emissions|CH4": (
                False,
                silicone.database_crunchers.QuantileRollingWindows,
            ),
            "Emissions|CO2|Biosphere": (
                False,
                silicone.database_crunchers.QuantileRollingWindows,
            ),
            "Emissions|CO2|Fossil": (
                False,
                silicone.database_crunchers.QuantileRollingWindows,
            ),
            "Emissions|CO": (
                False,
                silicone.database_crunchers.QuantileRollingWindows,
            ),
            "Emissions|N2O": (
                False,
                silicone.database_crunchers.QuantileRollingWindows,
            ),
            "Emissions|NH3": (
                False,
                silicone.database_crunchers.QuantileRollingWindows,
            ),
            "Emissions|NOx": (
                False,
                silicone.database_crunchers.QuantileRollingWindows,
            ),
            "Emissions|OC": (
                False,
                silicone.database_crunchers.QuantileRollingWindows,
            ),
            "Emissions|SOx": (
                False,
                silicone.database_crunchers.QuantileRollingWindows,
            ),
            "Emissions|NMVOC": (
                False,
                silicone.database_crunchers.QuantileRollingWindows,
            ),
            "Emissions|HFC134a": (
                False,
                silicone.database_crunchers.RMSClosest,
            ),
            "Emissions|HFC143a": (
                False,
                silicone.database_crunchers.RMSClosest,
            ),
            "Emissions|HFC227ea": (
                False,
                silicone.database_crunchers.RMSClosest,
            ),
            "Emissions|HFC23": (
                False,
                silicone.database_crunchers.RMSClosest,
            ),
            "Emissions|HFC32": (
                False,
                silicone.database_crunchers.RMSClosest,
            ),
            "Emissions|HFC4310mee": (
                False,
                silicone.database_crunchers.RMSClosest,
            ),
            "Emissions|HFC125": (
                False,
                silicone.database_crunchers.RMSClosest,
            ),
            "Emissions|SF6": (
                False,
                silicone.database_crunchers.RMSClosest,
            ),
            "Emissions|CF4": (
                False,
                silicone.database_crunchers.RMSClosest,
            ),
            "Emissions|C2F6": (
                False,
                silicone.database_crunchers.RMSClosest,
            ),
            "Emissions|C6F14": (
                False,
                silicone.database_crunchers.RMSClosest,
            ),
            "Emissions|CCl4": (
                True,
                silicone.database_crunchers.RMSClosest,
            ),
            "Emissions|CFC11": (
                True,
                silicone.database_crunchers.RMSClosest,
            ),
            "Emissions|CFC113": (
                True,
                silicone.database_crunchers.RMSClosest,
            ),
            "Emissions|CFC114": (
                True,
                silicone.database_crunchers.RMSClosest,
            ),
            "Emissions|CFC115": (
                True,
                silicone.database_crunchers.RMSClosest,
            ),
            "Emissions|CFC12": (
                True,
                silicone.database_crunchers.RMSClosest,
            ),
            "Emissions|CH2Cl2": (
                True,
                silicone.database_crunchers.RMSClosest,
            ),
            "Emissions|CH3Br": (
                True,
                silicone.database_crunchers.RMSClosest,
            ),
            "Emissions|CH3CCl3": (
                True,
                silicone.database_crunchers.RMSClosest,
            ),
            "Emissions|CH3Cl": (
                True,
                silicone.database_crunchers.RMSClosest,
            ),
            "Emissions|CHCl3": (
                True,
                silicone.database_crunchers.RMSClosest,
            ),
            "Emissions|HCFC141b": (
                True,
                silicone.database_crunchers.RMSClosest,
            ),
            "Emissions|HCFC142b": (
                True,
                silicone.database_crunchers.RMSClosest,
            ),
            "Emissions|HCFC22": (
                True,
                silicone.database_crunchers.RMSClosest,
            ),
            "Emissions|HFC152a": (
                True,
                silicone.database_crunchers.RMSClosest,
            ),
            "Emissions|HFC236fa": (
                True,
                silicone.database_crunchers.RMSClosest,
            ),
            "Emissions|HFC365mfc": (
                True,
                silicone.database_crunchers.RMSClosest,
            ),
            "Emissions|Halon1202": (
                True,
                silicone.database_crunchers.RMSClosest,
            ),
            "Emissions|Halon1211": (
                True,
                silicone.database_crunchers.RMSClosest,
            ),
            "Emissions|Halon1301": (
                True,
                silicone.database_crunchers.RMSClosest,
            ),
            "Emissions|Halon2402": (
                True,
                silicone.database_crunchers.RMSClosest,
            ),
            "Emissions|NF3": (
                True,
                silicone.database_crunchers.RMSClosest,
            ),
            "Emissions|C3F8": (
                True,
                silicone.database_crunchers.RMSClosest,
            ),
            "Emissions|C4F10": (
                True,
                silicone.database_crunchers.RMSClosest,
            ),
            "Emissions|C5F12": (
                True,
                silicone.database_crunchers.RMSClosest,
            ),
            "Emissions|C7F16": (
                True,
                silicone.database_crunchers.RMSClosest,
            ),
            "Emissions|C8F18": (
                True,
                silicone.database_crunchers.RMSClosest,
            ),
            "Emissions|cC4F8": (
                True,
                silicone.database_crunchers.RMSClosest,
            ),
            "Emissions|SO2F2": (
                True,
                silicone.database_crunchers.RMSClosest,
            ),
        }
        """
        Definition of our default set of variables to infill and how to infill them

        Each key is a variable we can infill.
        Each value is a tuple with the following:

        - `False` if we should use the 'full' infilling database,
          `True` if we should use the database
          that has information about CFCs and other species
          not typically modelled by IAMs
        - The database cruncher from silicone to use for infilling the variable
        """

        if variables_to_infill is None:
            # Technically, having to infill some of these
            # would have caused a scenario to fail vetting,
            # but we could at least in theory infill all of them.
            variables_to_infill = tuple(VARS_DB_CRUNCHERS.keys())

        lead_options = (
            ("Emissions|CO2",),
            ("Emissions|CO2|Fossil",),
        )

        infillers = {}

        for v_infill in variables_to_infill:
            cfcs, cruncher = VARS_DB_CRUNCHERS[v_infill]

            infillers[v_infill] = partial(
                do_ar6_like_infilling,
                follower=v_infill,
                lead_options=lead_options,
                db_file=ar6_infilling_db_file
                if not cfcs
                else ar6_infilling_db_cfcs_file,
                cruncher=cruncher,
                cfcs=cfcs,
            )

        # CO2 Energy and Industrial special case
        infillers["Emissions|CO2|Fossil"] = partial(
            do_ar6_like_infilling,
            follower="Emissions|CO2|Fossil",
            lead_options=(("Emissions|CO2",),),
            db_file=ar6_infilling_db_file,
            cruncher=VARS_DB_CRUNCHERS["Emissions|CO2|Fossil"][1],
            cfcs=False,
        )

        return cls(
            infillers=infillers,
            run_checks=run_checks,
            historical_emissions=historical_emissions,
            harmonisation_year=harmonisation_year,
            progress=progress,
            n_processes=n_processes,
        )



"""
Helpers for unit handling
"""

from __future__ import annotations

from collections.abc import Collection
from typing import TYPE_CHECKING

import pandas as pd
from pandas_openscm.index_manipulation import (
    set_index_levels_func,
)
from pandas_openscm.indexing import multi_index_match

from gcages.exceptions import MissingOptionalDependencyError

if TYPE_CHECKING:
    import pint


def assert_has_no_pint_incompatible_characters(
    units: Collection[str], pint_incompatible_characters: Collection[str] = {"-"}
) -> None:
    """
    Assert that a collection does not contain pint-incompatible characters

    Parameters
    ----------
    units
        Collection to check

        This is named `units` because we are normally checking collections of units

    pint_incompatible_characters
        Characters which are incompatible with pint

        You should not need to change this, but it is made an argument just in case

    Raises
    ------
    AssertionError
        `units` has elements that contain pint-incompatible characters
    """
    unit_contains_pint_incompatible = [
        u for u in units if any(pi in u for pi in pint_incompatible_characters)
    ]
    if unit_contains_pint_incompatible:
        msg = (
            "The following units contain pint incompatible characters: "
            f"{unit_contains_pint_incompatible=}. "
            f"{pint_incompatible_characters=}"
        )
        raise AssertionError(msg)


def strip_pint_incompatible_characters_from_unit_string(unit_str: str) -> str:
    """
    Strip pint-incompatible characters from a unit string

    Parameters
    ----------
    unit_str
        Unit string from which to strip pint-incompatible characters

    Returns
    -------
    :
        `unit_str` with pint-incompatible characters removed
    """
    return unit_str.replace("-", "")


def strip_pint_incompatible_characters_from_units(
    indf: pd.DataFrame, units_index_level: str = "unit"
) -> pd.DataFrame:
    """
    Strip pint-incompatible characters from units

    Parameters
    ----------
    indf
        Input data from which to strip pint-incompatible characters

    units_index_level
        Column in `indf`'s index that holds the units values

    Returns
    -------
    :
        `indf` with pint-incompatible characters
        removed from the `units_index_level` of its index.
    """
    res = indf.copy()
    res.index = res.index.remove_unused_levels()  # type: ignore # not in pandas-stubs
    res.index = res.index.set_levels(  # type: ignore # pandas-stubs out of date
        res.index.levels[res.index.names.index(units_index_level)].map(  # type: ignore # pandas-stubs out of date
            strip_pint_incompatible_characters_from_unit_string
        ),
        level=units_index_level,
    )

    return res


# TODO: move to pandas-openscm
def convert_unit_like(
    df: pd.DataFrame,
    target: pd.DataFrame,
    df_unit_level: str = "unit",
    target_unit_level: str | None = None,
    ur: pint.UnitRegistry | None = None,
) -> pd.DataFrame:
    """
    Convert to units like a different [pd.DataFrame][pandas.DataFrame]

    Parameters
    ----------
    df
        [pd.DataFrame][pandas.DataFrame] of which to convert the units

    target
        [pd.DataFrame][pandas.DataFrame] whose units the result should match

    df_unit_level
        Level in `df`'s index which has unit information

    target_unit_level
        Level in `target`'s index which has unit information

        If not provided, we assume this is the same as `df_unit_level`

    ur
        Unit registry to use for determining unit conversions

        If not provided, we use [openscm_units.unit_registry][]

    Returns
    -------
    :
        `df` with units that match `target`

    Raises
    ------
    MissingOptionalDependencyError
        `ur` is `None` and openscm-units is not installed
    """
    if target_unit_level is None:
        target_unit_col_use = df_unit_level
    else:
        target_unit_col_use = target_unit_level

    extra_index_levels_target = target.index.names.difference(df.index.names)  # type: ignore # pandas-stubs confused
    if extra_index_levels_target:
        msg = (
            "Haven't worked out the logic "
            "when the target has index levels which aren't in `df`. "
            f"{extra_index_levels_target=}"
        )
        raise NotImplementedError(msg)

    df_units_s = df.index.get_level_values(df_unit_level).to_series(
        index=df.index.droplevel(df_unit_level), name="df_unit"
    )
    target_units_s = target.index.get_level_values(target_unit_col_use).to_series(
        index=target.index.droplevel(target_unit_col_use), name="target_unit"
    )
    unit_map = pd.DataFrame([*target_units_s.align(df_units_s)]).T

    if (unit_map["df_unit"] == unit_map["target_unit"]).all():
        # Already in matching units
        return df

    if ur is None:
        try:
            import openscm_units

            ur = openscm_units.unit_registry
        except ImportError:
            raise MissingOptionalDependencyError(  # noqa: TRY003
                "convert_unit_like(..., ur=None, ...)", "openscm_units"
            )

    df_converted = df.reset_index(df_unit_level, drop=True)
    for (df_unit, target_unit), conversion_df in unit_map.groupby(
        ["df_unit", "target_unit"]
    ):
        conversion_factor = ur(df_unit).to(target_unit).m
        to_alter_loc = multi_index_match(df_converted.index, conversion_df.index)  # type: ignore
        df_converted.loc[to_alter_loc, :] *= conversion_factor

    # All conversions done so can simply assign the unit column.
    # When moving to pandas-openscm, check this carefully
    # using different input row order etc.

    unit_map_reordered = unit_map.reorder_levels(df_converted.index.names)
    res = set_index_levels_func(
        df_converted,
        {df_unit_level: unit_map_reordered["target_unit"].loc[df_converted.index]},
    ).reorder_levels(df.index.names)

    return res



"""
Internal consistency checking helpers
"""

# TODO: move this to pandas-openscm
from __future__ import annotations

from typing import TYPE_CHECKING

import pandas as pd

if TYPE_CHECKING:
    from gcages.typing import PINT_SCALAR


class InternalConsistencyError(ValueError):
    """
    Raised when there is an internal consistency issue in the data

    Specifically, the sum of components doesn't match some total
    """

    def __init__(
        self,
        differences: pd.DataFrame,
        data_that_was_summed: pd.DataFrame,
        tolerances: dict[str, float | PINT_SCALAR],
    ) -> None:
        differences_variables = differences.index.get_level_values("variable").unique()
        data_that_was_summed_relevant_for_differences = data_that_was_summed[
            data_that_was_summed.index.get_level_values("variable").map(
                lambda x: any(v in x for v in differences_variables)
            )
        ].index.to_frame(index=False)

        error_msg = (
            "Summing the components does not equal the total "
            f"when using the following tolerances: {tolerances}. "
            f"Differences:\n{differences}\n"
            "This is the data we used in the sum:\n"
            f"{data_that_was_summed_relevant_for_differences}"
        )

        super().__init__(error_msg)



"""
Definition of the pre-processor class
"""

from __future__ import annotations

import multiprocessing
from collections import defaultdict
from functools import partial
from typing import Protocol

import pandas as pd
from attrs import asdict, define
from pandas_openscm.index_manipulation import update_index_levels_func
from pandas_openscm.indexing import multi_index_lookup, multi_index_match
from pandas_openscm.parallelisation import ParallelOpConfig, apply_op_parallel_progress

from gcages.aggregation import get_region_sector_sum
from gcages.assertions import (
    assert_data_is_all_numeric,
    assert_has_index_levels,
    assert_index_is_multiindex,
    assert_only_working_on_variable_unit_region_variations,
)
from gcages.cmip7_scenariomip.gridding_emissions import (
    CO2_BIOSPHERE_SECTORS_GRIDDING,
    CO2_FOSSIL_SECTORS_GRIDDING,
    get_complete_gridding_index,
    to_global_workflow_emissions,
)
from gcages.cmip7_scenariomip.pre_processing.reaggregation import (
    ReaggregatorBasic,
    ToCompleteResult,
)
from gcages.completeness import NotCompleteError, assert_all_groups_are_complete
from gcages.index_manipulation import split_sectors
from gcages.renaming import SupportedNamingConventions, convert_variable_name
from gcages.testing import assert_frame_equal
from gcages.units_helpers import strip_pint_incompatible_characters_from_units


@define
class CMIP7ScenarioMIPPreProcessingResult:
    """
    Result of pre-processing with [CMIP7ScenarioMIPPreProcessor][(m).]

    This has more components than normal,
    because we need to support both the 'normal' global path
    and harmonising at the region-sector level.
    """

    assumed_zero_emissions: pd.DataFrame | None
    """
    Emissions that were asssumed to be zero during the processing
    """

    gridding_workflow_emissions: pd.DataFrame
    """
    Emissions that can be used with the gridding workflow
    """

    global_workflow_emissions: pd.DataFrame
    """
    Emissions that can be used with the 'normal' global workflow
    """

    global_workflow_emissions_raw_names: pd.DataFrame
    """
    Emissions consistent with those that can be used with the 'normal' global workflow

    The difference is that these are reported with CMIP7 ScenarioMIP naming,
    which isn't compatible with our SCM runners (for example),
    so is probably not what you want to use,
    but perhaps helpful for plotting and direct comparisons.
    """


def guess_reaggregator(
    indf: pd.DataFrame,
    region_level: str,
) -> ReaggregatorLike:
    """
    Guess the re-aggregator to use with a given dataset

    Parameters
    ----------
    indf
        Data for which to guess the re-aggregator

    region_level
        Region level in the data index

    Returns
    -------
    :
        Guessed re-aggregator

    Raises
    ------
    ValueError
        Re-aggregator could not be guessed for `indf`
    """
    assumed_model_regions = tuple(
        r for r in indf.index.get_level_values(region_level).unique() if r != "World"
    )
    errors_l = []
    for guess_cls in (ReaggregatorBasic,):
        guess = guess_cls(
            model_regions=assumed_model_regions, region_level=region_level
        )

        try:
            guess.assert_has_all_required_timeseries(indf)

        except NotCompleteError as exc:
            # Not a match
            errors_l.append(f"For {guess_cls}, error was:\n{exc}")
            continue

        else:
            return guess

    errors = "\n".join(errors_l)
    msg = (
        "Could not guess the reaggregator for the given input:\n"
        f"{indf}.\n"
        f"Errors:\n{errors}"
    )
    raise ValueError(msg)


def do_pre_processing(  # noqa: PLR0912, PLR0913, PLR0915
    indf: pd.DataFrame,
    reaggregator: ReaggregatorLike | None,
    time_name: str,
    run_checks: bool,
    world_gridding_sectors: tuple[str, ...] = ("Aircraft", "International Shipping"),
    table: str = "Emissions",
    level_separator: str = "|",
    co2_fossil_sectors: tuple[str, ...] = CO2_FOSSIL_SECTORS_GRIDDING,
    co2_biosphere_sectors: tuple[str, ...] = CO2_BIOSPHERE_SECTORS_GRIDDING,
    co2_name: str = "CO2",
) -> CMIP7ScenarioMIPPreProcessingResult:
    """
    Do the pre-processing for a given scenario

    This only works on a single scenario at a time,
    to make verification and processing simpler.

    Parameters
    ----------
    indf
        Input data to process

    reaggregator
        Re-aggregator to use during the processing

    time_name
        Name of the time axis in `indf`

    run_checks
        Should checks be run during the processing?

        If you know what you're doing, you can turn these off for speed.

    world_gridding_sectors
        Sectors that should only be gridded at the world level

    table
        Name of the 'table' for emissions

        Used to process and create variable names

    level_separator
        Separator between levels in the variable names

    co2_fossil_sectors
        Sectors to assume have an origin in fossil CO2 reservoirs

        These should be provided in the gridding naming convention

    co2_biosphere_sectors
        Sectors to assume have an origin in biospheric CO2 reservoirs

        These should be provided in the gridding naming convention

    co2_name
        String that indicates emissions of CO2 in variable names

    Returns
    -------
    :
        Results of the pre-processing
    """
    assert_only_working_on_variable_unit_region_variations(indf)

    if reaggregator is None:
        # Levels we will guess
        region_level = "region"
        unit_level = "unit"
        variable_level = "variable"

    else:
        region_level = reaggregator.region_level
        unit_level = reaggregator.unit_level
        variable_level = reaggregator.variable_level

    if run_checks:
        assert_has_index_levels(
            indf,
            ["model", "scenario", region_level, unit_level, variable_level],
        )

    if reaggregator is None:
        reaggregator = guess_reaggregator(indf, region_level=region_level)

    indf_reported_times = indf.dropna(how="all", axis="columns")

    if run_checks:
        indf_reported_times_nan = indf_reported_times.isnull().any(axis="columns")
        if indf_reported_times_nan.any():
            issue_rows = indf.loc[indf_reported_times_nan, :]
            msg = f"NaNs after dropping unreported times:\n{issue_rows}"
            raise AssertionError(msg)

    indf_clean_units = strip_pint_incompatible_characters_from_units(
        indf_reported_times,
        units_index_level=reaggregator.unit_level,
    )

    if run_checks:
        reaggregator.assert_has_all_required_timeseries(indf_clean_units)
        reaggregator.assert_is_internally_consistent(indf_clean_units)

    to_complete_result = reaggregator.to_complete(indf_clean_units)
    gridding_workflow_emissions = reaggregator.to_gridding_sectors(
        to_complete_result.complete
    )

    if run_checks:
        if gridding_workflow_emissions.isnull().any().any():
            msg = "NaN in `gridding_workflow_emissions`"
            raise AssertionError(msg)

        if gridding_workflow_emissions.columns.dtype != indf.columns.dtype:
            msg = "Column type does not match input"
            raise AssertionError(msg)

        complete_index_gridding = get_complete_gridding_index(
            model_regions=reaggregator.model_regions,
            world_gridding_sectors=world_gridding_sectors,
            world_region=reaggregator.world_region,
            region_level=reaggregator.region_level,
            variable_level=reaggregator.variable_level,
            table=table,
            level_separator=level_separator,
        )
        assert_all_groups_are_complete(
            gridding_workflow_emissions, complete_index=complete_index_gridding
        )

        # Check we didn't lose any mass
        grss = partial(
            get_region_sector_sum,
            region_level=reaggregator.region_level,
            world_region=reaggregator.world_region,
        )
        gridded_emisssions_sectoral_regional_sum = grss(gridding_workflow_emissions)

        in_emissions_totals_to_compare_to = multi_index_lookup(
            grss(
                # Make sure we only sum across the levels
                # that are useful for getting the total
                multi_index_lookup(
                    indf, reaggregator.get_internal_consistency_checking_index()
                )
            ),
            gridded_emisssions_sectoral_regional_sum.index,  # type: ignore # need to cast first or something
        )
        # No tolerance as this should be exact
        assert_frame_equal(
            gridded_emisssions_sectoral_regional_sum,
            in_emissions_totals_to_compare_to,
        )

    # Figure out the global workflow emissions
    global_workflow_emissions_from_gridding_emissions = to_global_workflow_emissions(
        gridding_workflow_emissions,
        time_name=time_name,
        region_level=reaggregator.region_level,
        world_region=reaggregator.world_region,
        # These have to be hard-coded to the IAM naming convention
        global_workflow_co2_fossil_sector="Energy and Industrial Processes",
        global_workflow_co2_biosphere_sector="AFOLU",
        co2_fossil_sectors=co2_fossil_sectors,
        co2_biosphere_sectors=co2_biosphere_sectors,
        co2_name=co2_name,
    )

    gwe_split = split_sectors(gridding_workflow_emissions, middle_level="species")
    species_from_gridding = tuple(gwe_split.index.get_level_values("species").unique())

    # Firstly drop out everything which was used for gridding
    indf_obviously_not_used_in_gridding = indf_clean_units.loc[
        ~multi_index_match(indf_clean_units.index, to_complete_result.complete.index)  # type: ignore
    ]

    # Now do the brute check on whatever is leftover
    def species_in_variable(variable: str, species: str, ls: str) -> bool:
        # ls: level separator
        # This mucking around is another illustration of the
        # issue with the data reporting format
        # (you need to check endswith and surrounded by the separator
        # to avoid accidental matches like VOC and OC)
        return variable.endswith(species) or (f"{ls}{species}{ls}" in variable)

    not_from_region_sector = [
        variable
        for variable in indf_obviously_not_used_in_gridding.index.get_level_values(
            variable_level
        ).unique()
        if not any(
            species_in_variable(variable, species=sg, ls=level_separator)
            for sg in species_from_gridding
        )
    ]
    global_workflow_emissions_not_from_gridding_emissions = indf_clean_units.loc[
        indf_clean_units.index.get_level_values(variable_level).isin(
            not_from_region_sector
        )
        # By definition, only want global emissions
        & (
            indf_clean_units.index.get_level_values(region_level)
            == reaggregator.world_region
        )
    ]
    # Don't report any carbon removal from the input
    # because it is already covered by the Emissions tree from the gridding timeseries
    global_workflow_emissions_not_from_gridding_emissions = global_workflow_emissions_not_from_gridding_emissions.loc[  # noqa: E501
        ~global_workflow_emissions_not_from_gridding_emissions.index.get_level_values(
            variable_level
        ).str.startswith("Carbon Removal")
    ]
    # Can't use these yet
    # TODO: implement support for baskets
    global_workflow_emissions_not_from_gridding_emissions = global_workflow_emissions_not_from_gridding_emissions.loc[  # noqa: E501
        ~global_workflow_emissions_not_from_gridding_emissions.index.get_level_values(
            unit_level
        ).str.contains("equiv")
    ]

    global_workflow_emissions_raw_names = pd.concat(
        [
            df.reorder_levels(
                global_workflow_emissions_from_gridding_emissions.index.names
            )
            for df in [
                global_workflow_emissions_from_gridding_emissions,
                global_workflow_emissions_not_from_gridding_emissions,
            ]
        ]
    )

    if run_checks:
        if global_workflow_emissions_raw_names.isnull().any().any():
            msg = "NaN in `global_workflow_emissions_raw_names`"
            raise AssertionError(msg)

        if global_workflow_emissions_raw_names.columns.dtype != indf.columns.dtype:
            msg = "Column type does not match input"
            raise AssertionError(msg)

    global_workflow_emissions = update_index_levels_func(
        global_workflow_emissions_raw_names,
        {
            "variable": partial(
                convert_variable_name,
                from_convention=SupportedNamingConventions.CMIP7_SCENARIOMIP,
                to_convention=SupportedNamingConventions.GCAGES,
            )
        },
    )

    res = CMIP7ScenarioMIPPreProcessingResult(
        assumed_zero_emissions=to_complete_result.assumed_zero,
        gridding_workflow_emissions=gridding_workflow_emissions,
        global_workflow_emissions=global_workflow_emissions,
        global_workflow_emissions_raw_names=global_workflow_emissions_raw_names,
    )

    return res


class ReaggregatorLike(Protocol):
    """
    Interface that can be used for re-aggregation
    """

    model_regions: tuple[str, ...]
    """Model regions to use while reaggregating"""

    region_level: str
    """Region level in the data index"""

    unit_level: str
    """Unit level in the data index"""

    variable_level: str
    """Variable level in the data index"""

    world_region: str
    """
    The value used when the data represents the sum over all regions

    (Having a value for this is odd,
    there should really just be no region level when data is the sum,
    but this is the data format used so we have to follow this convention.)
    """

    def assert_has_all_required_timeseries(self, indf: pd.DataFrame) -> None:
        """
        Assert that the data has all the required timeseries

        Parameters
        ----------
        indf
            Data to check

        Raises
        ------
        NotCompleteError
            `indf` is not complete
        """

    def assert_is_internally_consistent(self, indf: pd.DataFrame) -> None:
        """
        Assert that the data is internally consistent

        Parameters
        ----------
        indf
            Data to check

        Raises
        ------
        InternalConsistencyError
            The data is not internally consistent
        """

    def get_internal_consistency_checking_index(self) -> pd.MultiIndex:
        """
        Get the index which selects only data relevant for checking internal consistency

        Returns
        -------
        :
            Internal consistency checking index
        """

    def to_complete(self, raw: pd.DataFrame) -> ToCompleteResult:
        """
        Convert the raw data to complete data

        Parameters
        ----------
        raw
            Raw data

        Returns
        -------
        :
            To complete result
        """

    def to_gridding_sectors(self, indf: pd.DataFrame) -> pd.DataFrame:
        """
        Re-aggregate data to the sectors used for gridding

        Parameters
        ----------
        indf
            Data to re-aggregate

        Returns
        -------
        :
            Data re-aggregated to the gridding sectors
        """


@define
class CMIP7ScenarioMIPPreProcessor:
    """
    Pre-processor for CMIP7's ScenarioMIP

    For more details of the logic, see [gcages.cmip7_scenariomip.pre_processing][].
    """

    reaggregator: ReaggregatorLike | None = None
    """
    Re-aggregator to use when converting raw data to gridding sectors

    If not supplied, we guess the re-aggregator during processing
    """

    run_checks: bool = True
    """
    If `True`, run checks on both input and output data

    If you are sure about your workflow,
    you can disable the checks to speed things up
    (but we don't recommend this unless you really
    are confident about what you're doing).
    """

    world_gridding_sectors: tuple[str, ...] = ("Aircraft", "International Shipping")
    """
    Sectors that are only used for gridding at the world (i.e. regional sum) level
    """

    co2_fossil_sectors: tuple[str, ...] = CO2_FOSSIL_SECTORS_GRIDDING
    """
    Gridding sectors that are assumed to come from the fossil CO2 reservoir
    """

    co2_biosphere_sectors: tuple[str, ...] = CO2_BIOSPHERE_SECTORS_GRIDDING
    """
    Gridding sectors that are assumed to come from the biosphere CO2 reservoir
    """

    co2_name: str = "CO2"
    """
    Name used for CO2 in variable names
    """

    table: str = "Emissions"
    """
    The value used for the top level of variable names
    """

    level_separator: str = "|"
    """
    The separator between levels in variable names
    """

    progress: bool = True
    """
    Should progress bars be shown?
    """

    n_processes: int | None = multiprocessing.cpu_count()
    """
    Number of processes to use for parallel processing.

    Set to `None` to process in serial.
    """

    def __call__(
        self, in_emissions: pd.DataFrame
    ) -> CMIP7ScenarioMIPPreProcessingResult:
        """
        Pre-process

        Parameters
        ----------
        in_emissions
            Emissions to pre-process

        Returns
        -------
        :
            Pre-processed emissions
        """
        if self.run_checks:
            assert_index_is_multiindex(in_emissions)
            assert_data_is_all_numeric(in_emissions)

            if in_emissions.columns.name != "year":
                msg = "The input emissions' column name should be 'year'"
                raise AssertionError(msg)

        res_g = apply_op_parallel_progress(
            func_to_call=do_pre_processing,
            reaggregator=self.reaggregator,
            time_name="year",
            run_checks=self.run_checks,
            world_gridding_sectors=self.world_gridding_sectors,
            table=self.table,
            level_separator=self.level_separator,
            co2_fossil_sectors=self.co2_fossil_sectors,
            co2_biosphere_sectors=self.co2_biosphere_sectors,
            co2_name=self.co2_name,
            iterable_input=(
                gdf for _, gdf in in_emissions.groupby(["model", "scenario"])
            ),
            parallel_op_config=ParallelOpConfig.from_user_facing(
                progress=self.progress,
                max_workers=self.n_processes,
            ),
        )

        res_d = defaultdict(list)
        for res_ms in res_g:
            for k, v in asdict(res_ms).items():
                if v is not None:
                    res_d[k].append(v)

        result_initialiser = {k: pd.concat(v) for k, v in res_d.items()}
        if "assumed_zero_emissions" not in result_initialiser:
            result_initialiser["assumed_zero_emissions"] = None

        res = CMIP7ScenarioMIPPreProcessingResult(**result_initialiser)

        return res



"""
Handling of gridding emissions
"""

from __future__ import annotations

import itertools
import sys
from typing import TYPE_CHECKING

import pandas as pd
from pandas_openscm.grouping import groupby_except

from gcages.index_manipulation import (
    combine_sectors,
    combine_species,
    set_new_single_value_levels,
    split_sectors,
)

if sys.version_info >= (3, 11):
    from enum import StrEnum
else:
    from backports.strenum import StrEnum

if TYPE_CHECKING:
    from gcages.typing import NP_FLOAT_OR_INT

COMPLETE_GRIDDING_SPECIES: tuple[str, ...] = (
    "CO2",
    "CH4",
    "N2O",
    "BC",
    "CO",
    "NH3",
    "OC",
    "NOx",
    "Sulfur",
    "VOC",
)
"""
Complete set of species for gridding
"""

COMPLETE_GRIDDING_SECTORS_EXCEPT_CDR: tuple[str, ...] = (
    "Agricultural Waste Burning",
    "Agriculture",
    "Aircraft",
    "Energy Sector",
    "Forest Burning",
    "Grassland Burning",
    "Industrial Sector",
    "International Shipping",
    "Peat Burning",
    "Residential Commercial Other",
    "Solvents Production and Application",
    "Transportation Sector",
    "Waste",
    "Other CDR",
)
"""
Complete set of sectors for gridding excluding CDR sectors
"""

COMPLETE_GRIDDING_SECTORS_CDR: tuple[str, ...] = (
    "BECCS",
    "Enhanced Weathering",
    "Direct Air Capture",
    "Ocean",
    "Biochar",
    "Soil Carbon Management",
)
"""
Complete set of sectors for gridding CDR sectors
"""


def get_complete_gridding_index(  # noqa: PLR0913
    model_regions: tuple[str, ...],
    world_gridding_sectors: tuple[str, ...] = (
        "Aircraft",
        "International Shipping",
    ),
    world_region: str = "World",
    region_level: str = "region",
    variable_level: str = "variable",
    table: str = "Emissions",
    level_separator: str = "|",
) -> pd.MultiIndex:
    """
    Get the index of complete gridding data

    Parameters
    ----------
    model_regions
        Model regions to use while reaggregating

    world_gridding_sectors
        Sectors that should only be gridded at the world level

    world_region
        The value used when the data represents the sum over all regions

    region_level
        Region level in the data index

    variable_level
        Variable level in the data index

    table
        Name of the 'table' for emissions

        Used to process and create variable names

    level_separator
        Separator between levels in the variable names

    Returns
    -------
    :
        Index of complete gridding data
    """
    complete_world_variables = [
        level_separator.join([table, species, sectors])
        for species, sectors in itertools.product(
            COMPLETE_GRIDDING_SPECIES, world_gridding_sectors
        )
    ]
    world_required = pd.MultiIndex.from_product(
        [complete_world_variables, [world_region]], names=[variable_level, region_level]
    )

    model_region_sectors_except_cdr = sorted(
        set(COMPLETE_GRIDDING_SECTORS_EXCEPT_CDR) - set(world_gridding_sectors)
    )
    complete_model_region_variables_except_cdr = [
        level_separator.join([table, species, sectors])
        for species, sectors in itertools.product(
            COMPLETE_GRIDDING_SPECIES, model_region_sectors_except_cdr
        )
    ]

    complete_model_region_variables_cdr = [
        level_separator.join([table, "CO2", sectors])
        for sectors in COMPLETE_GRIDDING_SECTORS_CDR
    ]

    model_region_required = pd.MultiIndex.from_product(
        [
            [
                *complete_model_region_variables_except_cdr,
                *complete_model_region_variables_cdr,
            ],
            model_regions,
        ],
        names=[variable_level, region_level],
    )

    res: pd.MultiIndex = world_required.append(model_region_required)  # type: ignore # pandas-stubs out of date

    return res


class SpatialResolutionOption(StrEnum):
    """Spatial resolution option"""

    WORLD = "world"
    """Data reported at the world (i.e. global) level"""

    MODEL_REGION = "model_region"
    """Data reported at the (IAM) model region level"""


CO2_FOSSIL_SECTORS_GRIDDING: tuple[str, ...] = (
    "Aircraft",
    "BECCS",
    "International Shipping",
    "Energy Sector",
    "Industrial Sector",
    "Other CDR",
    "Enhanced Weathering",
    "Direct Air Capture",
    "Ocean",
    "Biochar",
    "Soil Carbon Management",
    "Residential Commercial Other",
    "Solvents Production and Application",
    "Transportation Sector",
    "Waste",
)
"""
Sectors that come from or go to fossil CO2 reservoirs (gridding naming convention)

BECCS is here because the carbon is stored permanently (or assumed to be).
It is grown then removed from the land pool,
so is 'net zero' from the land pool's point of view
(and handling this really well requires running a carbon cycle model
to determine the possible uptake from the BECCS land-use,
which isn't how the split between modelling domains works at the moment).

There is the same issue for some non-land CDR e.g. ocean alkalinity stuff.
Again, a handling sophisticiated enough to capture this properly
is beyond the scope of the fossil/biosphere split we're making here.

Not a perfect split with [CO2_BIOSPHERE_SECTORS_GRIDDING][(m).],
but the best we can do.
"""

CO2_BIOSPHERE_SECTORS_GRIDDING: tuple[str, ...] = (
    # Agriculture in biosphere because most of its emissions
    # are land carbon cycle (but not all, probably, in reality)
    "Agriculture",
    "Agricultural Waste Burning",
    "Forest Burning",
    "Grassland Burning",
    "Peat Burning",
)
"""
Sectors that come from biospheric CO2 reservoirs (gridding naming convention)

Not a perfect split with [CO2_FOSSIL_SECTORS_GRIDDING][(m).],
but the best we can do.
"""


def to_global_workflow_emissions(  # noqa: PLR0913
    gridding_emissions: pd.DataFrame,
    time_name: str = "year",
    region_level: str = "region",
    world_region: str = "World",
    global_workflow_co2_fossil_sector: str = "Fossil",
    global_workflow_co2_biosphere_sector: str = "Biosphere",
    co2_fossil_sectors: tuple[str, ...] = CO2_FOSSIL_SECTORS_GRIDDING,
    co2_biosphere_sectors: tuple[str, ...] = CO2_BIOSPHERE_SECTORS_GRIDDING,
    sectors_level: str = "sectors",
    species_level: str = "species",
    co2_name: str = "CO2",
) -> pd.DataFrame:
    """
    Convert gridding emissions to global workflow emissions

    Parameters
    ----------
    gridding_emissions
        Gridding emissions

    time_name
        Name of the time axis in `gridding_emissions`

    region_level
        Region level in the data index

    world_region
        The value used when the data represents the sum over all regions

    global_workflow_co2_fossil_sector
        Name of the CO2 'sector' with fossil origins to use in the output

    global_workflow_co2_biosphere_sector
        Name of the CO2 'sector' with biospheric origins to use in the output

    co2_fossil_sectors
        Sectors to assume have an origin in fossil CO2 reservoirs

    co2_biosphere_sectors
        Sectors to assume have an origin in biospheric CO2 reservoirs

    sectors_level
        Sectors level in the data index

    species_level
        Species level in the data index

    co2_name
        String that indicates emissions of CO2 in variable names

    Returns
    -------
    :
        Global workflow emissions
    """
    stacked: pd.DataFrame = (
        split_sectors(  # type: ignore
            gridding_emissions,
            middle_level=species_level,
            bottom_level=sectors_level,
        )
        .stack()
        .unstack("sectors")
    )

    world_locator = stacked.index.get_level_values(region_level) == world_region
    region_sector_df = stacked.loc[~world_locator]
    sector_df = stacked.loc[world_locator].reset_index("region", drop=True)

    gw_sector_df, gw_total_df = to_global_workflow_emissions_from_stacked(
        region_sector_df=region_sector_df,
        sector_df=sector_df,
        time_name=time_name,
        region_level=region_level,
        global_workflow_co2_fossil_sector=global_workflow_co2_fossil_sector,
        global_workflow_co2_biosphere_sector=global_workflow_co2_biosphere_sector,
        co2_fossil_sectors=co2_fossil_sectors,
        co2_biosphere_sectors=co2_biosphere_sectors,
        sectors_level=sectors_level,
        species_level=species_level,
        co2_name=co2_name,
    )

    gw_sector_df_input_like = set_new_single_value_levels(
        combine_sectors(
            gw_sector_df,  # type: ignore # fix when moving to pandas-openscm
            middle_level=species_level,
            bottom_level=sectors_level,
        ),
        {region_level: world_region},
    ).unstack(time_name)
    gw_total_df_input_like = set_new_single_value_levels(
        combine_species(gw_total_df, bottom_level=species_level),  # type: ignore # fix when moving to pandas-openscm
        {region_level: world_region},
    ).unstack(time_name)

    res = pd.concat(
        [
            df.reorder_levels(gridding_emissions.index.names)
            for df in [gw_total_df_input_like, gw_sector_df_input_like]
        ]
    )
    return res


def to_global_workflow_emissions_from_stacked(  # noqa: PLR0913
    region_sector_df: pd.DataFrame,
    sector_df: pd.DataFrame,
    time_name: str,
    region_level: str,
    global_workflow_co2_fossil_sector: str,
    global_workflow_co2_biosphere_sector: str,
    co2_fossil_sectors: tuple[str, ...],
    co2_biosphere_sectors: tuple[str, ...],
    sectors_level: str,
    species_level: str,
    co2_name: str,
) -> tuple[pd.Series[NP_FLOAT_OR_INT], pd.Series[NP_FLOAT_OR_INT]]:  # type: ignore # pandas-stubs out of date
    """
    Convert pre-stacked gridding emissions to global workflow emissions

    Parameters
    ----------
    region_sector_df
        Data with region and sector levels

    sector_df
        Data with sector levels only

    time_name
        Name of the time axis in `gridding_emissions`

    region_level
        Region level in the data index

    global_workflow_co2_fossil_sector
        Name of the CO2 'sector' with fossil origins to use in the output

    global_workflow_co2_biosphere_sector
        Name of the CO2 'sector' with biospheric origins to use in the output

    co2_fossil_sectors
        Sectors to assume have an origin in fossil CO2 reservoirs

    co2_biosphere_sectors
        Sectors to assume have an origin in biospheric CO2 reservoirs

    sectors_level
        Sectors level in the data index

    species_level
        Species level in the data index

    co2_name
        String that indicates emissions of CO2 in variable names

    Returns
    -------
    sectors
        Global workflow emissions with a sector level

    totals
        Global workflow emissions only with totals (no region or sector level)
    """
    region_sector_df_region_sum = groupby_except(region_sector_df, region_level).sum()

    sector_df_full = pd.concat([sector_df, region_sector_df_region_sum], axis="columns")

    co2_locator = (sector_df_full.index.get_level_values(species_level) == co2_name) & (
        sector_df_full.index.get_level_values("table") == "Emissions"
    )

    non_co2: pd.Series[NP_FLOAT_OR_INT] = sector_df_full[~co2_locator].sum(  # type: ignore # pandas-stubs out of date
        axis="columns"
    )

    not_used_cols = sorted(
        set(sector_df_full.columns)
        - {
            *co2_biosphere_sectors,
            *co2_fossil_sectors,
        }
    )
    if not_used_cols:
        msg = (
            "For the given inputs, not all CO2 sectors will be used.\n"
            f"{not_used_cols=}\n"
            f"{co2_fossil_sectors=}\n"
            f"{co2_biosphere_sectors=}\n"
        )
        raise AssertionError(msg)

    co2_fossil = set_new_single_value_levels(
        sector_df_full.loc[co2_locator, list(co2_fossil_sectors)].sum(axis="columns"),
        {sectors_level: global_workflow_co2_fossil_sector},
    )
    co2_biosphere = set_new_single_value_levels(
        sector_df_full.loc[co2_locator, list(co2_biosphere_sectors)].sum(
            axis="columns"
        ),
        {sectors_level: global_workflow_co2_biosphere_sector},
    )

    totals = non_co2
    sectors: pd.Series[NP_FLOAT_OR_INT] = pd.concat(  # type: ignore # pandas-stubs out of date
        [
            df.reorder_levels(co2_fossil.index.names)
            for df in [co2_fossil, co2_biosphere]
        ]
    )

    return sectors, totals



"""
Common components used across different re-aggregation strategies
"""

from __future__ import annotations

import pandas as pd
from attrs import define


@define
class ToCompleteResult:
    """
    Result of calling `to_complete` on a reaggregator
    """

    complete: pd.DataFrame
    """Complete [pd.DataFrame][pandas.DataFrame]"""

    assumed_zero: pd.DataFrame | None
    """
    The timeseries that were assumed to be zero to make `self.complete`

    If `None`, no timeseries were assumed to be zero.
    """
